; -------------------------------------------------------------------------------------
;
;  _____ __ __ ___ ___      _____       _   _        _____         _   _      ___ ___
; |__   |  |  | . |_  |    |_   _|___ _| |_| |_ _   |   __|___ ___| |_| |_ __|_  | . |
; |   __|-   -| . |_| |_     | | | . | . | . | | |  |   __| . |  _|  _|   |__| | |__ |
; |_____|__|__|___|_____|    |_| |___|___|___|_  |  |__|  |___|_| |_| |_|_|    |_|___|
;                                           |___|
;                                                                         Version 1.14
;
; -------------------------------------------------------------------------------------
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
; -------------------------------------------------------------------------------------
;
; A Direct-Threaded Forth for the ZX81 microcomputer
; 16 bit cell, 8 bit char (32 to 160), 8 bit (byte) adrs unit
;    Z80 BC = Forth TOS (top Param Stack item)
;        HL =       W    working register
;        DE =       IP   Interpreter Pointer
;        SP =       PSP  Param Stack Pointer
;     (RSP) =       RSP  Return Stack Pointer
;      (UP) =       UP   User Pointer
;        IY =       NEXT address
;    A, alternate register set = temporaries
;
;
; Memory map:
;
;   2000h             Forth kernel
;   3C00h             Character definition table (1024 bytes)
;   4000h             System variables
;   405Dh             User area (40 bytes)
;   407Dh             BASIC lines (for kernel load)
;   40ABh             Dfile (start with a HALT code)
;   43C6h             The FORTH word
;   ? h (DP)          Forth dictionary (user RAM, start at 43DBh)
;   ? h (DP+38h)      PAD buffer
;   7AFDh             Start of parameter stack (S0), grows down
;                     and start of Terminal Input Buffer (TIB), grows up
;   7BFDh             Start of return stack (R0), grows down
;
;
; Header structure:
;
; +---------------+
; |               |
; |-    link     -|     Link - points to the previous word's Length byte.
; |               |
; +---+-+---------+     P - Precedence bit, equals 1 for an IMMEDIATE word.
; |1|P|S| length  |     S - Smudge bit, used to prevent FIND from finding this word.
; +-+-+-+---------+     bits 0 to 4 are the word length
; |               |
; |-    name     -|
; |               |
; +---------------+
; |               |
; |-   cfa/pfa   -|
; |               |
; +---------------+
;
;
; The symbols include:
;
; addr   Cells  memory address
;   b      1      8 bit byte (i.e. hi 8 bits zero)
;   c      1      8 bit ascii character (hi 8 bits zero)
;   n      1      16 bit signed integer number
;   u      1      16 bit unsigned integer
;   x      1      Non-specific single cell
;  addr    1      Memory address
;   d      2      32 bit signed double integer, most significant portion with sign
;                 on top of stack.
;   ud     2      32 bit unsigned integer
;   xd     2      Non-specific 32 bit number
;   f      1      Boolean flag. 0=false, non-zero=true
;   ff     1      Boolean false flag=0
;   tf     1      Boolean true flag=non-zero
;
;
; When there is not provided UDG support, the characters that don't
; exist in ZX81 are mapped as follows:
;
;  char  ASCII CODE      char    ZX CODE    HOW TO TYPE IT
;    !       21      ->   [$]       8D        shift 1
;    @       40      ->   [£]       8C        shift 2
;    #       23      ->   [+]       95        shift 3
;    %       25      ->   [;]       99        shift 4
;    [       5B      ->   [(]       90        shift Q
;    ]       5D      ->   [)]       91        shift W
;    &       26      ->   [*]       97        shift E
;    ©       7F      ->   [,]       9A           -
;    '       27      ->   [?]       8F        shift T
;    _       5F      ->   [-]       96        shift Y
;    ~       7E      ->   [=]       94        shift A
;    |       7C      ->   [:]       8E        shift S
;    \       5C      ->   [/]       98        shift D
;    {       7B      ->   [<]       93        shift F
;    }       7D      ->   [>]       92        shift G
;    ^       5E      ->   ["]       8B        shift H
;

; -------------------------------------------------------------------------------------
; This code is intended for the sjasmplus, so compile it with:
;
;   sjasmplus.exe --lst TF79.asm
;
; and the files TF79.BIN, TF79.P and TF79.lst will be automatically generated in
; current directory.
;
;
; -------------------------------------------------------------------------------------
; Macros to define Forth headers
; HEAD  label,length,name,action
; IMMED label,length,name,action
;    label  = assembler name for this word
;             (special characters not allowed)
;    length = length of name field
;    name   = Forth's name for this word
;    action = code routine for this word, e.g.
;             DOCOLON, or DOCODE for code words
; IMMED defines a header for an IMMEDIATE word.
;
docode  equ 0
link    defl 0

head    macro label,length,name,action
        dw link
.lnk    equ $
        db length+$80,name
label:
        if action != docode
        call    action
        endif
link    defl .lnk
        endm

immed   macro label,length,name,action
        dw link
.lnk    equ $
        db length+$C0,name
label:
        if action != docode
        call    action
        endif
link    defl .lnk
        endm

nohead  macro label,action
label:
        if action != docode
        call    action
        endif
        endm

; The NEXT macro (7 bytes) assembles the 'next'
; code in-line in a Z80 CODE word.
next    macro
        ex      de,hl           ; The heart of FORTH! The NEXT engine
        ld      e,(hl)
        inc     hl
        ld      d,(hl)
        inc     hl
        ex      de,hl
        jp      (hl)
        endm

; NEXTHL is used when the IP is already in HL.
nexthl  macro
        ld      e,(hl)
        inc     hl
        ld      d,(hl)
        inc     hl
        ex      de,hl
        jp      (hl)
        endm

; -------------------------------------------------------------------------------------

; some useful ROM routines
;
KSCAN           equ     $02bb
DECODEKEY       equ     $07bd

        output TF79.BIN


;---------------------
; THE FORTH CORE CODE
;---------------------

        org $2000


; Pixel masks for high resolution graphics.
; Must be aligned to 256 byte boundary.
;
PIXMASK db $80,$40,$20,$10,$08,$04,$02,$01

; ---

; -----------------------
; THE 'WRITE AY' ROUTINE
; -----------------------
; write a byte at an AY register
; in: B = register number, C = data byte
;
writeay:
        ld      a,b
        out     ($df),a
        ld      a,c
        out     ($0f),a
        ret

; -----------------
;  Find Next Entry
;
findnext:
        ld      bc,$6007        ; next entry
        or      $ff
        out     (c),a
        call    api_responder
        cp      $3f             ; done?
        ret     z

        call    errorhandler
        ret


; ----------------------------------------
;  ASCII to ZX character conversion table
; ----------------------------------------
; Must be aligned to 256 byte boundary.
;
ASCTAB:
  db $00,$8D,$0B,$95,$0D,$99,$97,$8F,$10,$11,$17,$15,$1A,$16,$1B,$18   ;32-47
  db $1C,$1D,$1E,$1F,$20,$21,$22,$23,$24,$25,$0E,$19,$13,$14,$12,$0F   ;48-63
  db $8C,$26,$27,$28,$29,$2A,$2B,$2C,$2D,$2E,$2F,$30,$31,$32,$33,$34   ;64-79
  db $35,$36,$37,$38,$39,$3A,$3B,$3C,$3D,$3E,$3F,$90,$98,$91,$8B,$96   ;80-95
  db $0C,$A6,$A7,$A8,$A9,$AA,$AB,$AC,$AD,$AE,$AF,$B0,$B1,$B2,$B3,$B4   ;96-111
  db $B5,$B6,$B7,$B8,$B9,$BA,$BB,$BC,$BD,$BE,$BF,$93,$8E,$92,$94,$9A   ;112-127
  db $80,$81,$82,$83,$84,$85,$86,$87,$07,$06,$05,$04,$03,$02,$01,$88   ;128-143
  db $89,$8a,$08,$09,$0a,$9b,$9c,$9d,$9e,$9f,$a0,$a1,$a2,$a3,$a4,$a5   ;144-159


; ****************
; ** KEY TABLES **
; ****************

; -------------------------------
; THE 'UNSHIFTED' CHARACTER CODES
; -------------------------------

KEYTAB:
        db    'Z','X','C','V'
        db    'A','S','D','F','G'
        db    'Q','W','E','R','T'
        db    '1','2','3','4','5'
        db    '0','9','8','7','6'
        db    'P','O','I','U','Y'
        db    $0d,'L','K','J','H'
        db    $20,'.','M','N','B'

; -----------------------------
; THE 'SHIFTED' CHARACTER CODES
; -----------------------------

K_SHIFT:
        db    ':',';','?','/'
        db    '~','|','\','{','}'
        db    '[',']','&','`',"'"
        db    '!','@','#','%'
        db    $01                 ; KEY LEFT
        db    $08                 ; BACKSPACE
        db    $04                 ; CAPS LOCK
        db    $03                 ; KEY RIGHT
        db    $07                 ; KEY UP
        db    $09                 ; KEY DOWN
        db    '"',')','(','$','_'
        db    $0c                 ; FUNCTION
        db    '=','+','-','^'
        db    $1b                 ; BREAK
        db    ',','>','<','*'

; ---

; The setfast routine
setfast:
        ld      hl,CDFLAG
        bit     7,(hl)          ; sv CDFLAG
        ret     z               ;

        halt                    ; Wait for Interrupt
        out     ($fd),a         ;
        res     7,(hl)          ; sv CDFLAG
        ret

; The setslow routine
setslow ld      hl,CDFLAG
        set     6,(hl)
slowfast:
        ld      hl,$403b        ; Address the system variable CDFLAG.
        ld      a,(hl)          ; Load value to the accumulator.
        rla                     ; rotate bit 6 to position 7.
        xor     (hl)            ; exclusive or with original bit 7.
        rla                     ; rotate result out to carry.
        ret     nc              ; return if both bits were the same.
        set     7,(hl)          ; Indicate SLOW mode - Compute and Display.

; -----------------------
; THE 'MAIN DISPLAY' LOOP
; -----------------------
; This routine is executed once for every frame displayed.

display1:
        push    af              ; *             save main registers
        push    bc              ; **
        push    de              ; ***
        push    hl              ; ****

        ld      hl,(FRAMES)     ; fetch two-byte system variable FRAMES.
        dec     hl              ; decrement frames counter.
        ld      (FRAMES),hl     ; update system variable FRAMES
        call    KSCAN           ; (740) routine KEYBOARD gets the key row in H and
                                ; the column in L. Reading the ports also starts
                                ; the TV frame synchronization pulse. (VSYNC)
        ld      (LAST_K),hl     ; (16) update LAST_K with new values.
        ld      b,477/13        ; 
        djnz    $               ;
        call    display3

; ---

xisr:                           ; where IX should be set
        ld      a,r
        ld      bc,$1901
        ld      a,$f5
        call    $2b5            ; DISPLAY-5
        call    display3

display2:
        call    display1

pr_cur  ld      a,(FTFLAGS)     ; ftflags - bit 7 flash cursor, 1=on
        ld      e,a             ; save it to test STOPON/OFF bit later
        rla
        jr      nc,QSTOP
        ld      hl,CTIMER       ; cursor inverts every 22 frames
        dec     (hl)
        jr      nz,QSTOP

        ld      (hl),16h        ; reset timer count
        ld      hl,(CURADD)     ; screen cursor address
        ld      a,(hl)
        cp      $ff             ; outside the screen?
        jr      z,QSTOP         ; jump if so
        xor     $80
        ld      (hl),a

QSTOP   bit     3,e             ; check the STOPON/OFF bit, if it is set then
        jr      z,retint        ; verify the SHIFT+SPACE keys
        ld      hl,(LAST_K)     ;
        ld      a,3             ; check vertical sectors 0 and 1
        and     h               ;
        jr      nz,retint       ;
        rl      l               ; check horizontal sector 7 for SPACE key
        jp      nc,RBRK         ;
        jr      retint

; ---

display3
        pop     ix
        ld      a,(MARGIN)      ;+37h   ;
        neg                     ;-37h   ; number blank lines (55 or 31) system variable MARGIN
        inc     a               ;-36h   ; (negated as at BIOS:029Eh)
        ex      af,af'                  ; terminate vertical retrace,
        out     ($fe),a                 ;/and enable NMI generator and terminate vert. retrace
retint: pop     hl
        pop     de
        pop     bc
        pop     af
        ret


; ---------------------------------------------------------------------------
;                           ZXpand Access Routines
; ---------------------------------------------------------------------------

api_responder  equ $1ff6        ; wait for long command to finish, A has
                                ; response code on return
api_fileop     equ $1ff8        ; DE = fname pointer, with optional start,length
                                ;  A = operation.: 0 = load, 1 = delete,
                                ;                  2 = rename, 80-ff = save
api_sendstring equ $1ffa        ; DE = high bit terminated string, terminating
                                ; zero will be sent


response40:
        call    api_responder

; ----------------------
;  ZXpand Error handler
;
errorhandler:
        cp      $40
        ret     z               ; operation OK

        ld      hl,.L01+5
        ld      e,-10h          ; does not consider error > $49
        add     a,e
        ld      (hl),a
        call    docolon
        dw      XDQUOTE
.L01    db      5,"ZXP: "
        dw      ABORT

; -------------------------
;  Open a file for reading
;
initfileoprd:
        xor   a                 ; read

; -------------------------
;  Open a file for writing
;
initfileop:
        call    setfast
        push    af
        call    api_sendstring  ; push the file path at (de) - sets up C for us
        pop     af
        ld      bc,8007h        ; a = 0 for read, 1 for write
        out     (c),a
        call    response40
        ret

; ---

prepbfrwrite:
        ld      bc,0007h
        ld      a,0ffh
        out     (c),a
        ld      b,40h
        jr      interwritedelay

prepbfrread:
        ld      bc,0007h
        xor     a
        out     (c),a
        ld      b,20h

   ; falls though into...

;; a small delay in between interface writes
;;
interwritedelay:
        ld      a,8
.L01    dec     a
        jr      nz,.L01
        ret

; ----------
;  Open DIR
;
; push high-bit-terminated string at (de) to the interface as zero-terminated
; string. if de = 0, then only terminating zero is pushed
; in: DE = path address string or DE = 0 to current dir
;
opendir:
        call    api_sendstring
        ld      bc,$6007        ; open dir
        ld      a,0
        out     (c),a
        call    response40
        ret

; ---------------------------------------------------------
;  Convert ASCII string to ZX81 high-bit-terminated string
;  in: BC = string address, first byte is length
;  out: DE = converted string address
;
fname2zx:
        ld      a,(bc)          ; a = string lenght
        ld      h,b             ; string address to HL
        ld      l,c             ;
        push    bc
        ld      b,a
.L01    inc     hl
        ld      a,(hl)          ; read a char
        ld      d,ASCTAB/256
        ld      e,a
        ld      a,(de)
        ld      (hl),a          ;
        djnz    .L01            ; loop to get next char
        set     7,(hl)
        pop     de              ; string address
        inc     de              ; bypass length byte
        ret

; ---

; ---------------
;  CLICK Routine
; ---------------
; produces the sound for the key pressed
;
CLICK   push    bc              ; save key and give a short click
        ld      bc,$080e        ; set volume for channel A
        call    writeay
        ld      bc,$07fe        ; enable noise in channel A
        call    writeay
        ld      b,50
        djnz    $
        ld      bc,$07ff        ; stop sound out
        call    writeay
        pop     bc              ; get saved key
        ret

; ---------------------------
;  Print a Character Routine
; ---------------------------
; in: A = character code
;
prt_a:  push    hl
        push    de              ; IP
        push    bc              ; TOS
        ld      c,a             ; char to TOS
        ld      de,.L01         ; set IP
        jp      EMIT            ; emit char to terminal
.L01    dw      .L02
.L02    pop     de              ; IP (TOS was POP'ed by EMIT)
        pop     hl
        ret

; -----------------------
;  Scan Keyboard Routine
; -----------------------
; out: A = key character code or null if no key press
;
SCAN    call    BOUNC           ; Read keyboard matrix
        ld      hl,LKEY
        ld      b,23            ; Value for delay of the first repeat
        or      a
        jr      z,SCAN1         ; There was no key pressed
        cp      (hl)            ; Jumps if it is equal to the last pressed key
        jr      z,SCAN3         ;

SCAN1   ld      (hl),a          ; Save the status of the last key
        inc     hl              ; REP

SCAN2   ld      (hl),b          ; REP
        ret                     ; Returns with key code in A

SCAN3   inc     hl              ; REP
        ld      b,(hl)          ; Decrement the repeat counter
        djnz    SCAN4

        dec     hl              ; LKEY
        ld      a,(hl)          ; When zero, sets the delay value to
        ld      b,2             ; the other repeats
        jr      SCAN1

SCAN4   xor     a
        jr      SCAN2

BOUNC   ld      b,30
BOUN0   push    bc
        call    KSCAN           ; Scan keyboard (795 T_cycles)
        pop     bc              ; Repeat 30 times to avoid keybounce effect
        djnz    BOUN0           ;
        ld      a,l             ; Returns if no key was pressed
        inc     a               ;
        ret     z               ;

        ld      b,h
        ld      c,l
        call    DECODEKEY       ; Otherwise, decode key

        ld      a,0             ; Returns in case of multiple keys
        ret     nc              ;
        ld      de,KEYTAB-$7e   ; Gets the code. of the pressed key
        add     hl,de           ;
        ld      a,(hl)          ;
        cp      $04             ; SHIFT-9 ?
        ld      hl,FTFLAGS
        jr      nz,BOUN2
        ld      a,(hl)          ; toggles caps flag
        xor     $04
        ld      (hl),a
BOUN1   ld      a,(LAST_K)
        inc     a
        jr      nz,BOUN1

BOUN2   bit     2,(hl)          ; caps active?
        ret     z
        cp      'A'
        ret     c
        cp      '['
        ret     nc
        add     $20
        ret

;-------------------
; SCROLL UP routine
;-------------------
;
SCROLL  ld      hl,$21          ; scroll display one line up
        ld      de,(D_FILE)
        add     hl,de
        ld      bc,$2f7
        ldir
        inc     de
        ld      h,d             ; now clear the last screen line
        ld      l,e
        ld      (CURADD),hl     ; update CURADD (column 0 of line 23)
        push    hl              ; save the screen position
        inc     de
        ld      (hl),c          ; BC is 0 after last ldir
        ld      c,$1f
        ldir
        pop     hl              ; restore CURADD
        ret


; -------------------------------
;  FORTH INTERNAL CODE FRAGMENTS
; -------------------------------

; ----------------------------------------------
; DOCOLON, entered by CALL ENTER to enter a new
; high-level thread (colon def'n.)
; (internal code fragment, not a Forth word)
;
docolon:
        ld      hl,(RSP)        ;push old IP on ret stack
        dec     hl
        ld      (hl),d
        dec     hl
        ld      (hl),e
        ld      (RSP),hl
        pop     hl              ;pfa -> IP
NEXTHL  nexthl

; ----------------------------------------------
; DODOES, code action of DOES> clause
; entered by       CALL fragment
;                  parameter field
;                       ...
;        fragment: CALL DODOES
;                  high-level thread
; Enters high-level thread with address of
; parameter field on top of stack.
;
dodoes  ld      hl,(RSP)
        dec     hl
        ld      (hl),d
        dec     hl
        ld      (hl),e
        ld      (RSP),hl
        pop     de              ;adrs of new thread -> IP
        pop     hl              ;adrs of parameter field
        push    bc              ;push old TOS onto stack
        ld      b,h
        ld      c,l
NEXT:   next

; DOUSER, code action of USER, entered by CALL DOUSER
;  -- a-addr
;
douser: pop     hl              ; Pop parameter field address from stack
        push    bc              ; push old TOS
        ld      c,(hl)          ; Load single byte offset into C.
        ld      b,0             ;
        ld      hl,(UP)         ; load USER pointer
        add     hl,bc           ; add it to offset
        ld      b,h             ; load result in TOS
        ld      c,l
        next

; -------------------------
;  Forth words definitions
; -------------------------

; -----------------
;  THE BRANCH WORD
; -----------------
;  ( -- )
;
; branch always
;
        nohead BRANCH,docode
        ld      a,(de)          ;get inline value -> IP
        ld      l,a
        inc     de
        ld      a,(de)
        ld      h,a
        nexthl

; ------------------
;  THE ?BRANCH WORD
; ------------------
;  ( f -- )
;
; branch if TOS zero
;
        nohead QBRANCH,docode
        ld      a,b
        or      c               ; test old TOS
        pop     bc              ; pop new TOS
        jr      z,BRANCH        ; if old TOS=0, branch
        inc     de              ; else skip inline value
        inc     de
        next

; ---------------
;  THE (DO) WORD
; ---------------
;  ( n1 n2 -- )
;
; The run-time proceedure compiled by DO which moves the loop control
; parameters to the return stack.
;
        nohead XDO,docode
        ex      de,hl
        ex      (sp),hl         ; save IP and get n1 in HL
        ex      de,hl           ; DE<--LIMIT (n1) / BC<--initial INDEX (n2)
        ld      hl,(RSP)        ; HL<--return stack pointer
        dec     hl
        ld      (hl),d
        dec     hl
        ld      (hl),e          ; (R2)<--LIMIT
        dec     hl
        ld      (hl),b
        dec     hl
        ld      (hl),c          ; (R1)<--initial INDEX
        ld      (RSP),hl        ; (RP)<--(RP)-4
        pop     hl              ; restore the saved IP
        pop     bc              ; pop new TOS
        jp      NEXTHL

;-----------------
; THE (LOOP) WORD
; ----------------
;  ( -- )
;
; Run-time proceedure compiled by LOOP which increases the index and
; test for loop completion.
;
        nohead XLOOP,docode
        push    bc              ; balance the stack
        ld      bc,1            ; INCR
.L01    ld      hl,(RSP)        ; ((HL))=INDEX
        ld      a,(hl)          ; INDEX<--INDEX+INCR
        add     a,c
        ld      (hl),a
        ld      c,a
        inc     hl
        ld      a,(hl)
        adc     a,b
        ld      (hl),a
        inc     hl              ; ((HL))=LIMIT
        inc     b
        dec     b
        ld      b,a             ; (BC)<--new INDEX
        jp      m,.L02          ; if INCR>0
        ld      a,c
        sub     (hl)            ; then (A)<--INDEX - LIMIT
        ld      a,b
        inc     hl
        sbc     a,(hl)
        jp      .L03
.L02    ld      a,(hl)          ; else (A)<--LIMIT - INDEX
        sub     c
        inc     hl
        ld      a,(hl)
        sbc     a,b
.L03    pop     bc              ; if (A)<0
        jp      m,BRANCH        ; then loop again
        inc     hl              ; else done
        ld      (RSP),hl        ; discard R1 & R2
        inc     de              ; skip branch offset
        inc     de
        next

; ------------------
;  THE (+LOOP) WORD
; ------------------
;  ( n -- )
;
; The run-time proceedure compiled by +LOOP, which increments the loop
; index by n and tests for loop completion.
;
        nohead XPLOOP,docode
        jp XLOOP.L01

; -------------
; THE LIT WORD
; -------------
;  ( -- n )
;
; fetch inline literal to stack
;
        nohead LIT,docode
        push    bc
        ld      a,(de)
        inc     de
        ld      c,a
        ld      a,(de)
        inc     de
        ld      b,a
        next

; ------------
;  THE I WORD
; ------------
;  ( -- n )
;
; Copy the loop index onto the data stack. May only be used in
; the form:
;      DO ... I ... LOOP   or
;      DO ... I ... +LOOP
;
        head II,1,"I",docode
        MODULE II
        push    bc              ; push old TOS
        ld      hl,(RSP)        ; Get return stack pointer
L01     ld      c,(hl)          ; (DE)<--(R1)
        inc     hl              ;
        ld      b,(hl)          ;
        next
        ENDMODULE

; ------------
;  THE J WORD
; ------------
;  ( -- n )
;
; Return the index of the next outer loop. May only be used
; within a nested DO-LOOP in the form:
;      DO ... DO ... J ... LOOP ... LOOP
;
        head JJ,1,"J",docode
        push    bc              ; push old TOS
        ld      hl,(RSP)        ; Get return stack pointer
        inc     hl              ; Skip inner loop values
        inc     hl              ;
        inc     hl              ;
        inc     hl              ;
        jp      II.L01

; ----------------
;  THE LEAVE WORD
; ----------------
;  ( -- )
;
; Force  termination  of a DO-LOOP at the next LOOP or +LOOP  by
; setting the loop limit equal to the current value of the index.
; The index itself remains unchanged, and execution proceeds
; normally until the loop terminating word is encountered.
;
        head LEAVE,5,"LEAVE",docode
        push    de              ; save IP
        ld      hl,(RSP)
        ld      e,(hl)
        inc     hl
        ld      d,(hl)          ; (DE)<--(R1)=INDEX
        inc     hl
        ld      (hl),e
        inc     hl
        ld      (hl),d          ; (R2)<--(DE)=LIMIT
        pop     hl              ; IP --> HL
        jp      NEXTHL

; ---------------
;  THE EXIT WORD
; ---------------
;  ( -- )
;
; When compiled within a colon-definition, terminate execution
; of that definition, at that point. May not be used within a
; DO...LOOP.
;
        head EXIT,4,"EXIT",docode
        ld      hl,(RSP)        ;pop old IP from ret stack
        ld      e,(hl)
        inc     hl
        ld      d,(hl)
        inc     hl
        ld      (RSP),hl
        next

; -----------------
; THE EXECUTE WORD
; -----------------
;  ( addr -- )
;
; Execute the dictionary entry whose compilation address is on
; the stack.
;
        head EXECUTE,7,"EXECUTE",docode
        ld      h,b
        ld      l,c
        pop     bc
        jp      (hl)

; ---------------
;  THE NOOP WORD
; ---------------
;  ( -- )
;
; no operation (do nothing).
;
        head NOOP,4,"NOOP",docode
        jp      (iy)

; -------------------
;  THE VARIABLE WORD
; -------------------
;  ( -- )
;
; A defining word executed in the form:
;     VARIABLE <name>
; to create a dictionary entry for <name> and allot two bytes for
; storage in the parameter field. The application must initialize
; the stored value. When <name> is later executed, it will place
; the storage address on the stack.
;
; : VARIABLE   CREATE 0 , ;
;
        head VARIABLE,8,"VARIABLE",docolon
        dw CREATE,ZERO,COMMA,EXIT

; DOVAR, code action of VARIABLE, entered by CALL
; DOCREATE, code action of newly created words
docreate:
dovar:     ; ( -- addr)
        pop     hl              ; parameter field address
        push    bc              ; push old TOS
        ld      b,h             ; pfa = variable's adrs -> TOS
        ld      c,l
        next

; -------------------
;  THE CONSTANT WORD
; -------------------
;  ( x -- )
;
; A defining word used in the form:
;    n CONSTANT <name>
; to create a dictionary entry for <name>, leaving n in its
; parameter  field. When <name> is later executed, n will be
; left on the stack.
;
; : CONSTANT   CREATE , DOES> (machine code fragment)
;
        head CONSTANT,8,"CONSTANT",docolon
        dw CREATE,COMMA,XSCCODE

; DOCON, code action of CONSTANT,
; entered by CALL DOCON
docon:  pop     hl              ; parameter field address
        push    bc              ; old TOS
        ld      c,(hl)          ; fetch contents of parameter
        inc     hl              ; field -> TOS
        ld      b,(hl)
        next

; ------------
;  THE 0 WORD
; ------------
;  (  -- n )
;
; A constant for number zero.
;
        head ZERO,1,"0",docon
        dw 0

; ------------
;  THE 1 WORD
; ------------
;  (  -- 1 )
;
; A constant for number one.
;
        head ONE,1,"1",docon
        dw 1

; ------------
;  THE 2 WORD
; ------------
;  (  -- 2 )
;
; A constant for number two.
;
        head TWO,1,"2",docon
        dw 2

; ------------
;  THE 3 WORD
; ------------
;  ( -- 3 )
;
; A constant for number three.
;
        head THREE,1,"3",docon
        dw 3

; ----------------
;  THE (KEY) WORD
; ----------------
;  ( -- c )
;
; Default code for KEY
;
        head PKEY,5,"(KEY)",docode
        MODULE PKEY
        push    bc
        push    de
        call    setslow
L01     call    SCAN
        or      a
        jr      nz,L03

        ld      de,L02          ; call the PAUSE routine while no key press
        jp      PAUSE           ;
L02     dw      L01             ;

L03     ld      b,0
        ld      c,a
        call    CLICK
        pop     de
        ld      hl,FTFLAGS
        bit     6,(hl)          ; FTFLAGS - FAST flag
        jp      nz,FAST_
        jp      (iy)
        ENDMODULE

; -----------------
;  THE (EMIT) WORD
; -----------------
;  ( c -- )
;
; Default code for EMIT
;
        head PEMIT,6,"(EMIT)",docode
        push    de              ; sv IP
        ld      hl,(CURADD)     ; cursor address
        ld      a,(hl)          ; check for dfile end mark
        cp      $ff             ;
        ld      a,c             ; char to A
        call    z,SCROLL        ; scroll the screen if necessary
        cp      $20             ; printable?
        jr      c,ctrlch

        ld      d,ASCTAB/256    ; convert ASCII to ZX code
        ld      e,a             ;
        ld      a,(de)          ;
        ld      e,a
        ld      a,(INVCHR)
        xor     e
        and     $bf             ; reset bit 6
prch    ld      (hl),a
        inc     hl
        ld      a,(hl)
        cp      $76
        jr      nz,updcur
        inc     hl
updcur: ld      (CURADD),hl     ; cursor address
        pop     de              ; IP
        pop     bc              ; new TOS
        next

ctrlch: cp      $08             ; Backspace ?
        jr      z,backspace     ; Jump if it is Backspace (shift 0)
        cp      $0d             ; NEWLINE ?
        jr      z,emit_cr       ; Jump if it is NEWLINE
        ld      de,FTFLAGS      ; To be used by 'con' and 'coff'
        cp      $0e             ; Cursor on
        jr      z,con
        cp      $0f             ; Cursor off
        jr      z,coff
        ld      a,$0f           ; '?'
        jr      prch

emit_cr:
        ld      a,$76
        ld      bc,$21
        cpir
        jr      updcur

backspace:
        ld      (hl),$00        ; clear cursor
        dec     hl
        ld      a,$76
        cp      (hl)
        jr      nz,chk_dfstart
        dec     hl
chk_dfstart:                    ; check if is at Dfile start
        or      a
        ex      de,hl
        ld      hl,(D_FILE)
        sbc     hl,de
        ex      de,hl
        jr      c,clr_pos       ; Jump if inside of screen
        inc     hl
        inc     hl
clr_pos:
        ld      (hl),$00        ; Clear the screen position
        jr      updcur

; Cursor On
con:    ld      (hl),$80        ; print cursor
        ex      de,hl
        set     7,(hl)          ; FTFLAGS - cursor on
        ex      de,hl
        jr      updcur

; Cursor Off
coff:   ex      de,hl
        res     7,(hl)          ; FTFLAGS - cursor off
        ex      de,hl
        xor     a
        ld      (hl),a          ; cursor position
        jr      updcur

; ---------------
;  THE (AT) WORD
; ---------------
;  ( n1 n2 -- )
;
; Default code for AT
;
        head PAT,4,"(AT)",docode
        MODULE PAT
        pop     hl              ; n1
        push    de              ; save IP
        ld      a,c
        ld      (PR_CC),a
        ld      d,h             ; save n1 to DE
        ld      e,l             ;
        ld      b,5
L01     sla     l               ; n1*32
        rl      h               ;
        djnz    L01             ;
        add     hl,de           ; n1*33
        add     hl,bc           ; n1*33+n2
        inc     hl              ; n1*33+n2+1
        ld      de,(D_FILE)     ; dfile
        add     hl,de           ; n1*33+n2+1+dfile
        ld      (CURADD),hl     ; save it at CURADD
        pop     de              ; restore IP
        pop     bc              ; new TOS
        jp      (iy)
        ENDMODULE

; ------------------
;  THE (PAGE) WORD
; ------------------
;  ( -- )
;
; Default code for PAGE
;
        head PPAGE,6,"(PAGE)",docode
        MODULE PPAGE
        push    bc
        ld      hl,(D_FILE)
        inc     hl
        ld      (CURADD),hl
        ld      bc,$0418
L01     ld      a,(hl)
        cp      $76
        jr      z,L02
        ld      (hl),$00
L02     inc     hl
        dec     c
        jr      nz,L01
        djnz    L01

        pop     bc
        jp      (iy)
        ENDMODULE

; --------------
;  THE KEY WORD
; --------------
;  ( -- c )
;
; Leave the ASCII value of the next available character from
; the current input device.
;
        head KEY,3,"KEY",docode
        ld      hl,PKEY
        jp      (hl)

; ----------------
;  THE INKEY WORD
; ----------------
;  ( -- c )
;
; Leave the ASCII code of key pressed or null if none key was
; pressed.
;
        head INKEY,5,"INKEY",docode
        push    bc
        push    de
        call    BOUNC
        ld      b,0
        ld      c,a
        pop     de
        jp      (iy)

; ---------------
;  THE EMIT WORD
; ---------------
;  ( c -- )
;
; Transmit character to the current output device.
;
        head EMIT,4,"EMIT",docode
        ld      hl,PEMIT
        jp      (hl)

; -------------
;  THE AT WORD
; -------------
;  ( n1 n2 -- )
;
; Sets print position to row n1 and column n2
;
; : AT   DUP 1F + 4038 C! SWAP 21 * + 400C @ 1+ + 400E ! ;
;
        head AT,2,"AT",docode
        ld      hl,PAT
        jp      (hl)

; ---------------
;  THE PAGE WORD
; ---------------
;  ( -- )
;
; Clear the screen
;
        head PAGE,4,"PAGE",docode
        ld      hl,PPAGE
        jp      (hl)

; --------------------
;  THE ?TERMINAL WORD
; --------------------
;  ( -- f )
;
; Perform a test of the terminal keyboard for actuation of the
; break key. A true flag indicates actuation.
;
        head QTERMINAL,9,"?TERMINAL",docode
        push    bc
        ld      a,$7f           ; read port $7FFE - keys B,N,M,.,SPACE.
        in      a,($fe)         ;
        rra                     ; carry will be set if space not pressed.
        ccf
        jp      ZLESS+2

; -----------------
;  THE STOPON WORD
; -----------------
;  ( -- )
;
; Enables the BREAK key
;
        head STOPON,6,'STOPON',docode
        ld      hl,FTFLAGS
        set     3,(hl)
        jp      (iy)

; ------------------
;  THE STOPOFF WORD
; ------------------
;  ( -- )
;
; Disables the BREAK key
;
        head STOPOFF,7,'STOPOFF',docode
        ld      hl,FTFLAGS
        res     3,(hl)
        jp      (iy)

; ------------------
;  THE CHR$128 WORD
; ------------------
;  ( f -- )
;
; Enable/disable the CHR$128 character generator mode.
;
        head CHR128,7,"CHR$128",docode
        ld      a,(IREG)
        rra
        rr      c
        rla
        ld      i,a
        jr      PLOT.L06

; ------------------
;  THE INVERSE WORD
; ------------------
;  ( -- )
;
; Invert the characters transmited to the output terminal
; ( use with CHR$128 mode disabled )
        head INVERSE,7,"INVERSE",docode
        ld      hl,INVCHR
        ld      a,(hl)
        xor     $80
        ld      (hl),a
        jp      (iy)

; ---------------
;  THE PLOT WORD
; ---------------
;  ( n1 n2 n3 -- )
;
; Plot at n1 (X), n2 (Y) with mode n3
; (0 = unplot, 1=plot, 2=move, 3=change)
;
        head PLOT,4,"PLOT",docode
        MODULE PLOT
        push    bc
        exx
        pop     bc              ; n
        pop     de              ; y
        ld      hl,$4037
        ld      (hl),e          ; YCOORD
        dec     hl
        ld      a,$2F           ; 47-y
        sub     e               ;
        pop     de              ; x
        ld      (hl),e          ; XCOORD
        jr      c,L05           ; y out of range, do nothing
        rra                     ; (47-y)/2
        ld      b,a
        rl      c               ; n*2+cy
        ld      a,$3f
        cp      e
        jr      c,L05           ; out of range
        srl     e               ; x/2
        rl      c               ; (n*2)*2
        ld      a,e
        inc     b               ; range 1 to 24
        ld      hl,(D_FILE)     ; Dfile
        ld      de,$21
        inc     hl
L01     add     hl,de           ; (47-y/2)*33
        djnz    L01             ;

        sbc     hl,de           ; adjust address
        ld      e,a             ; (47-y/2)*33+x/2
        add     hl,de           ;
        ld      a,(hl)          ; get char at plot position
        rlca
        cp      $10             ; is it a graphic char?
        ld      a,(hl)
        jr      c,L02
        xor     a               ; clear it if isn't
L02     ld      e,a             ; sv screen char
        ld      d,$87
        ld      a,c             ; mode & pixel quadrant to A
        cpl
        and     $03             ; select complemented pixel quadrant
        ld      b,a             ; reserve it
        jr      z,L03           ; jump if it is quadrant 3 pixel
        cpl                     ; calc pixel char mask
        add     a,$02           ;
        adc     a,$03           ;
        ld      d,a             ; sv mask at D
        ld      b,e             ; sv screen char at B
L03     ld      a,c             ; get mode
        rrca                    ;
        rrca                    ;
        rrca                    ;
        sbc     a,a             ; 0 for unplot/move, $ff for plot/change
        bit     3,c             ; check for move/change mode
        jr      nz,L04          ; jump if it is

; plot/unplot mode
        xor     e
        rlca
        sbc     a,a
        xor     b
L04     and     d
        xor     e
        ld      (hl),a
L05     exx
L06     pop     bc
        jp      (iy)
        ENDMODULE

; ----------------
;  THE SOUND WORD
; ----------------
;  ( c n -- )
; Writes data c to register n of the AY sound chip.
;
        head SOUND,5,"SOUND",docode
        ld      a,c             ; save AY register
        pop     bc              ; data byte to C
        ld      b,a             ; AY register to B
        call    writeay
        pop     bc              ; new TOS
        jp      (iy)

; ---------------
;  THE BELL WORD
; ---------------
;  ( -- )
; Activate  a terminal bell or noise-maker as appropriate to the
; device in use (emits a 1318Hz sound for ~40ms).
;
; : BELL  7 EMIT ;
;
        head BELL,4,"BELL",docode
        MODULE BELL
        push    bc
        ld      bc,$0070        ; set channel A freq
        call    writeay
        ld      c,b
        inc     b
        call    writeay
        ld      bc,$07fe        ; enable tone to channel A
        call    writeay
        ld      c,b             ; set channel A vol to 7
        inc     b
        call    writeay
        ld      hl,$0a00        ; beep duration
L01     dec     hl
        ld      a,h
        or      l
        jr      nz,L01

        ld      c,a             ; reset channel A freq
        call    writeay+1
        pop     bc
        jp      (iy)
        ENDMODULE

; ---------------
;  THE SLOW WORD
; ---------------
;  ( -- )
;
; Set the SLOW mode
;
        head SLOW_,4,"SLOW",docode
        ld      hl,FTFLAGS
        res     6,(hl)          ; FTFLAGS - FAST flag
        call    setslow
        jp      (iy)

; ---------------
;  THE FAST WORD
; ---------------
;  ( -- )
;
; Set the FAST mode
;
        head FAST_,4,"FAST",docode
        ld      hl,FTFLAGS
        set     6,(hl)          ; FTFLAGS - FAST flag
        call    setfast
        res     6,(hl)
        jp      (iy)

; ----------------
;  THE PAUSE WORD
; ----------------
;  ( -- )
;  Switch to the next task in multi-task environment.
;
        head PAUSE,5,"PAUSE",docode
        ld      hl,NOOP
        jp      (hl)

; ---------------
;  THE WAIT WORD
; ---------------
;  ( n -- )
;
; Waits until n timer interrupts have occurred. While waiting, execute the word PAUSE.
;
        head WAIT,4,"WAIT",docode
        MODULE WAIT
        push    de              ; sv IP
L01     ld      hl,FRAMES       ; current lsb frame counter to A
        ld      a,(hl)          ;
        ld      e,a             ; sv it to E
L02     xor     (hl)            ; wait until next frame
        ld      a,e             ;
        jr      z,L02           ;
        ld      de,L03          ; set IP to return from PAUSE
        jp      PAUSE           ; execute PAUSE routine
L03     dw      L04
L04     ld      a,b
        or      c
        dec     bc              ; decrement counter
        jr      nz,L01
        pop     de              ; sv IP
        pop     bc              ; new TOS
        jp      (iy)
        ENDMODULE

; --------------
;  THE P@  WORD
; --------------
;  ( addr -- b )
;
; Read byte from Z80 input port adr
;
        head PFETCH,2,"P@",docode
        in      c,(c)        ; read port (BC) to C
        ld      b,0
        jp      (iy)

; --------------
;  THE P!  WORD
; --------------
;  ( b addr --  )
;
; Write byte to Z80 output port adr
;
        head PSTORE,2,"P!",docode
        pop     hl          ; char in L
        out     (c),l       ; to port (BC)
        pop     bc          ; pop new TOS
        jp      (iy)

; --------------
;  THE SP@ WORD
; --------------
;  ( -- addr )
;
; Get the data stack pointer
;
        head SPFETCH,3,"SP@",docode
        push    bc
        ld      hl,0
        add     hl,sp
        ld      b,h
        ld      c,l
        jp      (iy)

; --------------
;  THE SP! WORD
; --------------
;  ( -- )
;
; Set the data stack pointer to initial value (S0).
;
        head SPSTORE,3,"SP!",docode
        ld      sp,(s0)
        pop     bc              ; adjust the stack
        jp      (iy)

; ---------------
;  THE RP@  WORD
; ---------------
;  ( -- addr )
;
; Get the return stack pointer
;
        head RPFETCH,3,"RP@",docode
        push    bc
        ld      bc,(RSP)
        jp      (iy)

; --------------
;  THE RP! WORD
; --------------
;  ( -- )
;
; Set the return stack pointer to initial value (R0).
;
        head RPSTORE,3,"RP!",docode
        ld      hl,(r0)
        ld      (RSP),hl
        jp      (iy)

; --------------
;  THE >R  WORD
; --------------
;  ( x -- )  ( R: -- x )                "to-r"
;
; Transfer n to the return stack. Every >R must be balanced by
; a R> in the same control structure nesting level of a colon-
; definition.
;
        head TOR,2,">R",docode
        ld      hl,(RSP)
        dec     hl
        ld      (hl),b
        dec     hl
        ld      (hl),c                  ;  (R1)<--(DE)
        ld      (RSP),hl                ;  (RP)<--(RP)-2
        pop     bc
        jp      (iy)

; -------------
;  THE R> WORD
; -------------
;  ( -- x )     R: x --                 "r-from"
;
; Transfer n from the return stack to the data stack.
;
        head FROMR,2,"R>",docode
        push    bc
        ld      hl,(RSP)
        ld      c,(hl)
        inc     hl
        ld      b,(hl)
        inc     hl
        ld      (RSP),hl
        jp      (iy)

; -------------
;  THE R@ WORD
; -------------
;  ( -- x )     R: x -- x               "r-fetch"
;
; Copy the number on top of the return stack to the data stack.
;
        head RFETCH,2,"R@",docode
        push    bc
        ld      hl,(RSP)
        ld      c,(hl)
        inc     hl
        ld      b,(hl)
        jp      (iy)

; --------------
;  THE DUP WORD
; --------------
;  ( x -- x x )                         "dupe"
;
; Leave a copy of the top stack number.
;
        head DUP,3,"DUP",docode
        push    bc
        jp      (iy)

; ---------------
;  THE ?DUP WORD
; ---------------
;  ( x -- 0 )   if zero                 "query-dupe"
;  ( x -- x x ) if nonzero
;
; Duplicate n if it is non-zero.
;
        head QDUP,4,"?DUP",docode
        ld      a,b
        or      c
        jr      nz,DUP
        jp      (iy)

; ---------------
;  THE DROP WORD
; ---------------
;  ( x -- )
; Drop the top number from the stack.
;
        head DROP,4,"DROP",docode
        pop     bc
        jp      (iy)

; ---------------
;  THE SWAP WORD
; ---------------
;  ( x1 x2 -- x2 x1 )
;
; Exchange the top two stack values.
;
        head SWAP,4,"SWAP",docode
        pop     hl
        push    bc
        ld      b,h
        ld      c,l
        jp      (iy)

; ---------------
;  THE OVER WORD
; ---------------
;  ( x1 x2 -- x1 x2 x1 )
;
; Leave a copy of the second number on the stack.
;
        head OVER,4,"OVER",docode
        pop     hl
        push    hl
        push    bc
        ld      b,h
        ld      c,l
        jp      (iy)

; ---------------
;  THE PICK WORD
; ---------------
;  ( x1 -- x2 )
;
; Return the contents of the n-th stack value , not counting n itself.
; An error condition results for n less than one.
;      1 PICK = DUP
;      2 PICK = OVER
;
; : PICK   1- 2* SP@ 2+ + @ ;
;
        head PICK,4,"PICK",docode
        dec     bc
        sla     c               ; 2*
        rl      b               ;
        jp      c,NEXT          ; do nothing if x1 < 1
        ld      h,b
        ld      l,c
        add     hl,sp
        ld      c,(hl)          ; get element to be PICK'ed
        inc     hl              ;
        ld      b,(hl)          ;
        jp      (iy)

; ---------------
;  THE ROLL WORD
; ---------------
;  ( x -- )
;
; Extract  the  n-th stack value to the top of  the  stack,  not
; counting  n  itself,  moving  the remaining  values  into  the
; vacated position.   An error condition results for n less than
; one.
;      3  ROLL  =  ROT
;      1  ROLL  =  null operation
;
        head ROLL,4,"ROLL",docode
        MODULE ROLL
        push    bc
        exx                     ;save IP
        pop     bc
        dec     bc
        sla     c               ;2*
        rl      b               ;
        jp      c,NEXT          ; do nothing if x1 < 1
        inc     bc              ;1+
        ld      h,b
        ld      l,c
        dec     bc              ;bytes count to be moved
        add     hl,sp
        push    hl              ;destination
        push    bc
        ld      d,(hl)          ;get element to be ROLL'ed
        dec     hl              ;
        ld      e,(hl)          ;
        dec     hl              ;HL = origin
        pop     bc
        ex      de,hl
        ex      (sp),hl
        ex      de,hl
        ld      a,b
        or      c               ; count=0?
        jr      z,L01           ;
        lddr
L01     exx                     ;restore IP
        pop     bc              ;get TOS
        pop     hl              ;adjust SP
        jp      (iy)
        ENDMODULE

; --------------
;  THE ROT WORD
; --------------
;  ( x1 x2 x3 -- x2 x3 x1 )             "rote"
;
; Rotate the top three values, bringing the deepest to the top.
;
        head ROT,3,"ROT",docode
        pop     hl
        ex      (sp),hl
        push    bc
        ld      b,h
        ld      c,l
        jp      (iy)

; ---------------
;  THE -ROT WORD
; ---------------
;  ( x1 x2 x3 -- x3 x1 x2 )             "minus rote"
;
; Moves the top of stack to the third position.
;
        head MINUSROT,4,'-ROT',docode
        ld      h,b
        ld      l,c
        pop     bc              ;b2
        ex      (sp),hl
        push    hl
        jp      (iy)

; ----------------
;  THE UNDER WORD
; ----------------
;  ( x1 x2 -- x2 )
;
; Drop the second item in the data stack.
;
        head UNDER,5,"UNDER",docode
        pop     hl
        jp      (iy)

; ---------------
;  THE 2DUP WORD
; ---------------
;  ( xd -- xd xd )                      "two-dupe"
;
; Duplicate the top double number on the stack.
;
; : 2DUP   OVER OVER ;
;
        head TWODUP,4,"2DUP",docode
        pop     hl
        push    hl
        push    bc
        push    hl
        jp      (iy)

; ----------------
;  THE 2DROP WORD
; ----------------
;  ( xd -- )                            "two-drop"
;
; Drop the top double number on the stack.
;
; : 2DROP   DROP DROP ;
;
        head TWODROP,5,"2DROP",docode
        pop     bc
        pop     bc
        jp      (iy)

; ----------------
;  THE 2SWAP WORD
; ----------------
;  ( xd1 xd2 -- xd2 xd1 )               "two-swap"
;
; Exchange the top two double numbers on the stack.
;
; : 2SWAP   ROT >R ROT R> ;
;
        head TWOSWAP,5,"2SWAP",docode
        pop     hl  ; bc=x4, hl=x3
        exx
        pop     bc
        pop     hl  ; bc'=x2, hl'=x1
        exx
        push    hl
        push    bc
        exx
        push    hl
        push    bc
        exx
        pop     bc
        jp      (iy)

; ----------------
;  THE 2OVER WORD
; ----------------
;  ( xd1 xd2 -- xd1 xd2 xd1 )           "two-over"
;
; Leave a copy of the second double number on the stack.
;
; : 2OVER   >R >R 2DUP R> R> 2SWAP ;
;
        head TWOOVER,5,"2OVER",docode
        exx
        pop     hl
        pop     de
        pop     bc
        push    bc
        push    de
        push    hl
        exx
        push    bc
        exx
        push    bc
        push    de
        exx
        pop     bc
        jp      (iy)

; ------------
;  THE @ WORD
; ------------
;  ( addr -- x )                        "fetch"
;
; Leave on the stack the number contained at addr.
;
        head FETCH,1,"@",docode
        ld      h,b
        ld      l,c
        ld      c,(hl)
        inc     hl
        ld      b,(hl)
        jp      (iy)

; -------------
;  THE C@ WORD
; -------------
;  ( addr -- b )                        "c-fetch"
;
; Leave on the stack the contents of the byte at addr (with
; higher bits zero, in a 16-bit field).
;
        head CFETCH,2,"C@",docode
        ld      a,(bc)
        ld      c,a
        ld      b,0
        jp      (iy)

; -------------
;  THE 2@ WORD
; -------------
;  ( addr -- x1 x2 )
;
; Fetch the two cells x1 and x2 from addr.
;
        head TWOFETCH,2,"2@",docode
        ld      h,b
        ld      l,c
        ld      c,(hl)
        inc     hl
        ld      b,(hl)
        inc     hl
        ld      a,(hl)
        inc     hl
        ld      h,(hl)
        ld      l,a
        push    hl
        jp      (iy)

; ------------
;  THE ! WORD
; ------------
;  ( x addr -- )                        "store"
;
; Store n at addr.
;
        head STORE,1,"!",docode
        ld      h,b
        ld      l,c
        pop     bc
        ld      (hl),c
        inc     hl
        ld      (hl),b
        pop     bc              ;new TOS
        jp      (iy)

; -------------
;  THE C! WORD
; -------------
;  ( x addr -- )                        "c-store"
;
; Store the least significant 8-bits of n at addr.
;
        head CSTORE,2,"C!",docode
        ld      h,b            ;addr in HL
        ld      l,c
        pop     bc             ;char in BC
        ld      (hl),c
        pop     bc             ;new TOS
        jp      (iy)

; -------------
;  THE 2! WORD
; -------------
;  ( x1 x2 addr -- )
;
; Store the two cells x1 and x2 at addr.
;
        head TWOSTORE,2,"2!",docode
        ld      h,b
        ld      l,c
        pop     bc
        ld      (hl),c
        inc     hl
        ld      (hl),b
        inc     hl
        pop     bc
        ld      (hl),c
        inc     hl
        ld      (hl),b
        pop     bc
        jp      (iy)

; -------------
;  THE +! WORD
; -------------
;  ( n addr -- )                        "plus-store"
;
; Add n to the 16-bit value at the address, by the convention
; given for + .
;
        head PLUSSTORE,2,"+!",docode
        pop     hl
        ld      a,(bc)
        add     a,l
        ld      (bc),a
        inc     bc
        ld      a,(bc)
        adc     a,h
        ld      (bc),a
        pop     bc              ;new TOS
        jp      (iy)

; ------------
;  THE + WORD
; ------------
;  ( n1 n2 -- n3 )                      "plus"
;
; Leave the arithmetic sum of n1 plus n2.
;
        head PLUS,1,"+",docode
        pop     hl
        add     hl,bc
        ld      b,h
        ld      c,l
        jp      (iy)

; ------------
;  THE - WORD
; ------------
;  ( n1 n2 -- n3 )                      "minus"
;
; Subtract n2 from n1 and leave the difference n3.
;
        head MINUS,1,"-",docode
        pop     hl
        and     a
        sbc     hl,bc
        ld      b,h
        ld      c,l
        jp      (iy)

; -------------
;  THE D+ WORD
; -------------
;  ( d1 d2 --- d3 )                     "d-plus"
;
; Leave the arithmetic sum of d1 plus d2.
;
        head DPLUS,2,"D+",docode
        exx
        pop     bc              ; BC'=d2lo
        exx
        pop     hl              ; HL=d1hi,BC=d2hi
        exx
        pop     hl              ; HL'=d1lo
        add     hl,bc
        push    hl              ; 2OS=d1lo+d2lo
        exx
        adc     hl,bc           ; HL=d1hi+d2hi+cy
        ld      b,h
        ld      c,l
        jp      (iy)

; -------------
;  THE D- WORD
; -------------
;  ( d1 d2 --- d3 )                     "d-plus"
;
; Leave the arithmetic sum of d1 plus d2.
;
        head DMINUS,2,"D-",docolon
        dw DNEGATE,DPLUS,EXIT

; -------------
;  THE U* WORD
; -------------
;  ( un1 un2 -- ud3 )                   "u-times"
;
; Perform an unsigned multiplication of un1 by un2,  leaving
; the double number product ud3.  All values are unsigned.
;
        head UTIMES,2,"U*",docode
        MODULE UTIMES
        pop     hl              ; u1 in HL
        push    de              ; save IP
        ex      de,hl           ; u1 to DE, u2 already in BC
        ld      hl,0            ; result will be in HLDE
        ld      a,17            ; loop counter
        or      a               ; clear cy
L01     rr      h
        rr      l
        rr      d
        rr      e
        jr      nc,L02
        add     hl,bc
L02     dec     a
        jr      nz,L01
        ld      b,h             ; hi result in BC
        ld      c,l             ;
        pop     hl              ; restore IP to HL
        push    de              ; lo result
        nexthl
        ENDMODULE

; ------------
;  THE * WORD
; ------------
;  ( n1 n2 -- n3 )                      "times"
;
; Leave the arithmetic product of n1 times n2.
;
; : * U* DROP ;
;
        head TIMES,1,"*",docode
        MODULE TIMES
        pop     hl              ; u1 in HL
        push    de              ; save IP
        ld      d,b             ; u2 in DE
        ld      e,c             ;
        ld      a,l             ;
        ld      c,h             ;
        ld      b,$10           ;
        ld      hl,0            ;
L01     srl     c               ;
        rra                     ; SRL CA  (MPCATOR)
        jr      nc,L02          ; LSB (CA)=0?
        add     hl,de           ; no, add mpcant to HL
L02     sla     e               ;
        rl      d               ; SLA DE  (MPCANT 2 *)
        djnz    L01             ; do all 16 bits
        ld      b,h             ; TOS<--product
        ld      c,l
        pop     hl              ; restore IP
        nexthl
        ENDMODULE

; ---------------
;  THE /MOD WORD
; ---------------
;  ( n1 n2 -- n3 n4 )                   "divide-mod"
;
; Divide  n1 by n2 and leave the remainder n3 and
; quotient  n4.  n3 has the same sign as n1.
;
        head DIVMOD,4,"/MOD",docode
        MODULE DIVMOD
        pop     hl              ; dividend
        push    de              ; Save IP
        ld      d,b             ; divisor
        ld      e,c             ;
        ld      a,d             ;
        or      e               ; div by 0?
        jr      nz,L01          ;
        ld      de,-1           ; yes: rem <-- -1,  quot <-- dividend
        jr      L07             ; exit
L01     xor     a               ; reset negate flag
        push    af
        bit     7,d             ; divisor negative?
        jr      z,L02           ;
        ld      a,e             ; yes
        cpl                     ;
        ld      e,a             ;
        ld      a,d             ;
        cpl                     ;
        ld      d,a             ;
        inc     de              ; (de)<--(de)'s 2's complement
        pop     af
        scf                     ; set negate flag
        push    af
L02     bit     7,h             ; dividend negative?
        jr      z,L03           ;
        ld      a,l             ; yes
        cpl                     ;
        ld      l,a             ;
        ld      a,h             ;
        cpl                     ;
        ld      h,a             ;
        inc     hl              ; (hl)<--(hl)'s 2's complement
        pop     af
        ccf                     ; negate flag
        inc     a               ; dividend sign flag
        push    af
L03     ld      a,l             ;
        ld      c,h             ; (ca)<--dividend
        ld      hl,0            ; prime remainder
        ld      b,10h           ; loop counter
L04     rla                     ;
        rl      c               ; rl ca
        adc     hl,hl           ; (hl)<--(hl) 2 * carry +
        sbc     hl,de           ; underflow?
        jr      nc,L05          ; no
        add     hl,de           ; yes, restore remainder
L05     ccf                     ;
        djnz    L04             ; do all 16 bits
        rla                     ;
        rl      c               ; rl ca
        ld      e,a             ;
        ld      d,c             ; (de)<--quotient
        pop     af
        jr      z,L06           ; dividend positive
        ld      a,l             ;
        cpl                     ;
        ld      l,a             ;
        ld      a,h             ;
        cpl                     ;
        ld      h,a             ;
        inc     hl              ; rem gets dividend's sign
L06     jr      nc,L07          ; quotient positive
        ld      a,e             ;
        cpl                     ;
        ld      e,a             ;
        ld      a,d             ;
        cpl                     ;
        ld      d,a             ;
        inc     de              ; negative quotient
L07     ld      b,d             ; quotient
        ld      c,e
        ex      (sp),hl         ; restore IP and push remainder
        nexthl
        ENDMODULE

; ------------
;  THE / WORD
; ------------
;  ( n1 n2 -- n3 )                      "divide"
;
; Divide  n1  by n2 and leave the quotient n3.
; n3  is  rounded toward zero.
;
; : /   /MOD UNDER ;
;
        head DIV,1,"/",docode
        MODULE DIV
        pop     hl              ; divider
        push    de              ; save IP
        push    hl              ; returns divider to the stack
        ld      de,L01          ; IP = star1
        jp      DIVMOD
L01     dw      L02
L02     pop     hl              ; DROP 2OS
        pop     hl              ; restore IP to HL
        nexthl
        ENDMODULE

; ----------------
;  THE U/MOD WORD
; ----------------
;  ( ud1 un2 -- un3 un4 )               "u-divide-mod"
;
; Perform  the  unsigned division of double number ud1 by un2,
; leaving the remainder un3,  and the quotient un4.
; All values are unsigned.
;
        head UDIVMOD,5,"U/MOD",docode
        MODULE UDIVMOD
        pop     hl              ; ud hi to HL
        ex      de,hl           ; ip to stack, ud lo to DE
        ex      (sp),hl         ;
        ex      de,hl           ; BC = divisor, HLDE = dividend
        ld      a,16            ; loop counter
        sla     e
        rl      d               ; hi bit DE -> carry
L01     adc     hl,hl           ; rot left w/ carry
        jr      nc,L02
        ; case 1: 17 bit, cy:HL = 1xxxx
        or      a               ; we know we can subtract
        sbc     hl,bc
        or      a               ; clear cy to indicate sub ok
        jr      L03
        ; case 2: 16 bit, cy:HL = 0xxxx
L02     sbc     hl,bc           ; try the subtract
        jr      nc,L03          ; if no cy, subtract ok
        add     hl,bc           ; else cancel the subtract
        scf                     ;   and set cy to indicate
L03     rl      e               ; rotate result bit into DE,
        rl      d               ; and next bit of DE into cy
        dec     a
        jr      nz,L01
        ; now have complemented quotient in DE,
        ; and remainder in HL
        ld      a,d
        cpl
        ld      b,a
        ld      a,e
        cpl
        ld      c,a
        pop     de
        push    hl
        jp      (iy)
        ENDMODULE

; -------------
;  THE 1+ WORD
; -------------
;  ( n -- n+1 )                         "one-plus"
;
; Increment n by one, according to the operation of + .
;
        head ONEPLUS,2,"1+",docode
        inc     bc
        jp      (iy)

; -------------
;  THE 2+ WORD
; -------------
;  ( n -- n+2 )                         "two-plus"
;
; Increment n by two, according to the operation of + .
;
        head TWOPLUS,2,"2+",docode
        inc     bc
        inc     bc
        jp      (iy)

; -------------
;  THE 1- WORD
; -------------
;  ( n -- n-1 )                         "one-minus"
;
; Decrement n by one, according to the operation of - .
;
        head ONEMINUS,2,"1-",docode
        dec     bc
        jp      (iy)

; -------------
;  THE 2- WORD
; -------------
;  (n -- n-1 )                          "two-minus"
;
; Decrement n by two, according to the operation of - .
;
        head TWOMINUS,2,"2-",docode
        dec     bc
        dec     bc
        jp      (iy)

; -------------
;  THE 2* WORD
; -------------
;  ( n1 -- n2 )                         "two-times"
;
; Leave 2*(n1).
;
        head TWOTIMES,2,"2*",docode
        sla     c
        rl      b
        jp      (iy)

; -------------
;  THE 2/ WORD
; -------------
;  ( n1 -- n2 )                         "two-divide"
;
; Leave (n1)/2.
;
        head TWODIV,2,"2/",docode
        sra     b
        rr      c
        jp      (iy)

; ---------------
;  THE RAND WORD
; ---------------
;  ( u -- )
; Set the random number seed.
;
        head RAND,4,"RAND",docode
        MODULE RAND
        ld      a,b
        or      c
        jr      nz,L01          ; use seed if provided
        ld      bc,(FRAMES)     ; else use FRAMES
L01     ld      (SEED),bc
        pop     bc
        jp      (iy)
        ENDMODULE

; --------------
;  THE RND WORD
; --------------
;  ( u1 -- u2 )
; Get random number between 0 and u1-1.
;
        head RND,3,"RND",docode
; -----------------------------------------------------------------------------
;  Generates 16-bit pseudorandom numbers
;  16-bit xorshift pseudorandom number generator by John Metcalf
;  (https://wikiti.brandonw.net/index.php?title=Z80_Routines:Math:Random)
;  Outputs: bc  Pseudorandom number
;  Using the xor-shift method:
;    bc ^= bc << 7
;    bc ^= bc >> 9
;    bc ^= bc << 8
;  Some alternative shift triplets which also perform well are:
;       {6, 7, 13}; {7, 9, 13}; {9, 7, 13}.
;
        push    bc
        ld      bc,(SEED)
        ld      a,b
        rra
        ld      a,c
        rra
        xor     b
        ld      b,a
        ld      a,c
        rra
        ld      a,b
        rra
        xor     c
        ld      c,a
        xor     b
        ld      b,a
        ld      (SEED),bc
        call     docolon
        dw UTIMES,UNDER,EXIT

; --------------
;  THE AND WORD
; --------------
;  ( x1 x2 -- x3 )
;
; Leave the bitwise logical 'and' of n1 and n2.
;
        head AND_,3,"AND",docode
        pop     hl
        ld      a,b
        and     h
        ld      b,a
        ld      a,c
        and     l
        ld      c,a
        jp      (iy)

; -------------
;  THE OR WORD
; -------------
;  ( x1 x2 -- x3 )
;
; Leave the bitwise inclusive-or of two numbers.
;
        head OR_,2,"OR",docode
        pop     hl
        ld      a,b
        or      h
        ld      b,a
        ld      a,c
        or      l
        ld      c,a
        jp      (iy)

; --------------
;  THE XOR WORD                         "x-or"
; --------------
;  ( x1 x2 -- x3 )
;
; Leave the bitwise exclusive-or of two numbers.
;
        head XOR_,3,"XOR",docode
        pop     hl
        ld      a,b
        xor     h
        ld      b,a
        ld      a,c
        xor     l
        ld      c,a
        jp      (iy)

; -----------------
;  THE TOGGLE WORD
; -----------------
;  ( addr c -- )
; Changes the byte stored at address addr, by exclusive-ORing it with c.
;
        head TOGGLE,6,"TOGGLE",docode
        pop     hl
        ld      a,c
        xor     (hl)
        ld      (hl),a
        pop     bc
        jp      (iy)

; -----------------
;  THE NEGATE WORD
; -----------------
;  ( n -- -n )
;
; Leave the two's complement of a number,  i.e.,  the difference
; of zero less n.
;
        head NEGATE,6,"NEGATE",docode
        ld      a,b
        cpl
        ld      b,a
        ld      a,c
        cpl
        ld      c,a
        inc     bc
        jp      (iy)

; ------------------
;  THE DNEGATE WORD
; ------------------
;   ( d -- -d )
;
; Leave the double number two's complement of a double  number,
; i.e., the difference 0 less d.
;
; : DNEGATE  SWAP INVERT SWAP INVERT 1. D+ ;
;
        head DNEGATE,7,"DNEGATE",docode
        pop     hl              ;(HL)<--d1L
        sub     a               ;(A)<--0
        sub     l
        ld      l,a             ;(L)<--NEG(L)
        ld      a,$00
        sbc     a,h
        ld      h,a             ;(H)<--NEG(H)
        ld      a,$00
        sbc     a,c
        ld      c,a             ;(C)<--NEG(C)
        ld      a,$00
        sbc     a,b
        ld      b,a             ;(B)<--NEG(B)
        push    hl
        jp      (iy)

; -------------
;  THE +- WORD
; -------------
;  ( n1 n2 -- n3 )
;
; Leave the two's complement of n1 if n2 is negative
;
; : +-   0< IF NEGATE THEN ;
;
        head PLUSMINUS,2,"+-",docode
        sla     b
        pop     bc
        jr      c,NEGATE
        jp      (iy)

; --------------
;  THE D+- WORD
; --------------
;  ( d1 n -- d2 )
;
;  Leave the two's complement of d1 if n is negative
;
; : D+-  0< IF DNEGATE THEN ;
;
        head DPLUSMINUS,3,"D+-",docode
        sla     b
        pop     bc
        jr      c,DNEGATE
        jp      (iy)

; -------------
;  THE 0= WORD
; -------------
;  ( n -- flag )                        "zero-equals"
;
; True if n is zero.
;
        head ZEQUAL,2,"0=",docode
        ld      a,b
        or      c
        ld      bc,0
        jp      nz,NEXT
        inc     c
        jp      (iy)

; -------------
;  THE 0< WORD
; -------------
;  ( n -- flag )                        "zero-less"
;
; True if n is less than zero (negative)
;
        head ZLESS,2,"0<",docode
        sla     b               ; sign bit -> cy flag
        ld      bc,0
        jp      nc,NEXT
        inc     c
        jp      (iy)

; -------------
;  THE 0> WORD
; -------------
;  ( n -- flag )                        "zero-greater"
;
; True if n is greater than zero.
;
        head ZGREATER,2,"0>",docode
        ld      hl,0
        jp      LESS+1

; ------------
;  THE = WORD
; ------------
;  ( n1 n2 -- flag )                    "equals"
;
; True if n1 is equal to n2.
;
        head EQUAL,1,"=",docode
        pop     hl
        or      a
        sbc     hl,bc       ; x1-x2 in HL, SZVC valid
        jr      z,TOSTRUE
TOSFALSE:
        ld      bc,0
        jp      (iy)

; ------------
;  THE < WORD
; ------------
;  ( n1 n2 -- flag )                    "less-than"
;
; True if n1 is less than n2.
;
        head LESS,1,"<",docode
        pop     hl
        or      a
        sbc     hl,bc       ; n1-n2 in HL, SZVC valid
; if result negative & not OV, n1<n2
; neg. & OV => n1 +ve, n2 -ve, rslt -ve, so n1>n2
; if result positive & not OV, n1>=n2
; pos. & OV => n1 -ve, n2 +ve, rslt +ve, so n1<n2
; thus OV reverses the sense of the sign bit
        jp      pe,REVSENSE  ; if OV, use rev. sense
        jp      p,TOSFALSE   ;   if +ve, result false
TOSTRUE:
        ld      bc,1         ;   if -ve, result true
        jp      (iy)
REVSENSE:
        jp      m,TOSFALSE   ; OV: if -ve, reslt false
        jr      TOSTRUE      ;     if +ve, result true

; ------------
;  THE > WORD
; ------------
;  ( n1 n2 -- flag )                    "greater-than"
;
; True if n1 is greater than n2.
;
        head GREATER,1,">",docode
        ld      h,b
        ld      l,c
        pop     bc
        jp      LESS+1

; -------------
;  THE U< WORD
; -------------
;  ( un1 un2 -- flag )                  "u-less-than"
;
; Leave the flag representing the magnitude comparison of un1 < un2
; where un1 and un2 are treated as 16-bit unsigned integers.
;
        head ULESS,2,"U<",docode
        pop     hl
        xor      a
        sbc     hl,bc   ; u1-u2 in HL, SZVC valid
        ld      b,a     ; ld bc,0
        ld      c,a
        jp      nc,NEXT ; false
        inc     c       ; true
        jp      (iy)

; -------------
;  THE D< WORD
; -------------
;  ( d1 d2 -- flag )                    "d-less-than"
;
; True if d1 is less than d2.
;
        head DLESS,2,"D<",docode
        MODULE DLESS
        exx
        pop     bc              ; d2lo
        pop     hl              ; d1hi
        ex      (sp),hl         ; get d1lo in HL', push d1hi
        exx
        pop     hl              ; d1hi, BC=d2hi
        and     a
        sbc     hl,bc           ; d1hi-d2hi
        jr      z,L01
        add     hl,bc           ; restore d1hi in HL
        jp      LESS+1

L01     push    de
        exx
        pop     de
        jp      LESS+1
        ENDMODULE

; --------------
;  THE NOT WORD
; --------------
;  ( f1 -- f2 )               165
;
; Reverse the boolean value of f1.  This is identical to 0=.
;
; : NOT  0= ;
;
        head NOT,3,'NOT',docode
        jp ZEQUAL

; --------------
;  THE ABS WORD
; --------------
;  ( n1 -- n2 )                         "absolute"
;
; Leave the absolute value of a number.
;
; : ABS   DUP +- ;
;
        head ABS,3,"ABS",docode
        push bc
        jp PLUSMINUS

; -------------
;  THE CR WORD
;  (  --  )                             "c-r"
;
; Cause a carriage-return and line-feed to occur at the current
; output device.
;
        head CR,2,"CR",docode
        push    bc
        ld      c,$0d
        jp      EMIT

; ----------------
;  THE CMOVE WORD
; ----------------
;  ( addr1 addr2 n -- )                 "c-move"
;
; Move n bytes beginning at address addr1 to addr2. The contents
; of addr1 is moved first proceeding toward high memory.  If n
; is zero nothing is moved.
;
        head CMOVE,5,"CMOVE",docode
        MODULE CMOVE
        pop     hl            ; destination adrs
        ex      de,hl         ; DE = 'dst', HL = IP
        ex      (sp),hl       ; HL = 'src', IP on stack
        ld      a,b           ; test for count=0
        or      c             ;
        jr      z,L01         ; yes
        ldir                  ; move 'n' bytes
L01     pop     hl            ; restore IP
        pop     bc            ; pop new TOS
        jp      NEXTHL
        ENDMODULE

; -----------------
;  THE <CMOVE WORD
; -----------------
;  ( addr1 addr2 u -- )                 "reverse-c-move"
;
; Copy  n bytes beginning at addr1 to addr2.   The move proceeds
; within the bytes from high memory toward low memory.
;
        head REVCMOVE,6,"<CMOVE",docode
        MODULE REVCMOVE
        pop hl          ; destination adrs
        ex de,hl        ; IP to HL
        ex (sp),hl      ; sv IP, source adrs to HL
        ex de,hl
        ld a,b          ; test for count=0
        or c
        jr z,L01
        add hl,bc       ; last byte in destination
        dec hl
        ex de,hl
        add hl,bc       ; last byte in source
        dec hl
        lddr            ; move from top to bottom
L01     pop de          ; IP
        pop bc          ; pop new TOS
        jp      (iy)
        ENDMODULE

; ---------------
;  THE MOVE WORD
; ---------------
;  ( addr1 addr2 n -- )
;
; Move the specified quantity n of 16-bit memory cells beginning
; at addr1 into memory at addr2.  The contents of addr1 is moved
; first.  If n is negative or zero, nothing is moved.
;
        head MOVE,4,"MOVE",docode
        sla c
        rl b
        jr CMOVE

; ---------------
;  THE FILL WORD
; ---------------
;  ( addr n byte -- )
;
; Fill memory beginning at address with a sequence of n copies of
; byte. If the quantity n is less than or equal to zero, take no
; action.
;
        head FILL,4,"FILL",docode
        MODULE FILL
        ld      a,c             ; character in a
        exx                     ; use alt. register set
        pop     bc              ; count in bc
        pop     de              ; address in de
        or      a               ; clear carry flag
        ld      hl,$ffff
        adc     hl,bc           ; test for count=0 or 1
        jp      m,L01           ; count<0, skip
        jr      nc,L01          ; no cy: count=0, skip
        ld      (de),a          ; fill first byte
        jr      z,L01           ; zero, count=1, done
        dec     bc              ; else adjust count,
        ld      h,d             ; let hl = start adrs,
        ld      l,e
        inc     de              ; let de = start adrs+1
        ldir                    ; copy (hl)->(de)
L01     exx                     ; back to main reg set
        pop     bc              ; pop new TOS
        jp      (iy)
        ENDMODULE

; ----------------
;  THE ERASE WORD
; ----------------
;  ( addr u -- )
;
; Fill an area of memory over n bytes with zeros, starting  at
; addr. If n is zero or less, take no action.
;
; : ERASE   0 FILL ;
;
        head ERASE,5,"ERASE",docode
        push    bc
        ld      bc,0
        jp      FILL

; -----------------
;  THE BLANKS WORD
; -----------------
;  ( addr u -- )
;
; Fill an area of memory over n bytes with the value for  ASCII
; blank,  starting at addr.  If n is less than or equal to zero,
; take no action.
;
; : BLANKS    BL FILL ;
;
        head BLANKS,6,"BLANKS",docode
        push    bc
        ld      bc,$20
        jp      FILL

; ----------------
;  THE DEPTH WORD
; ----------------
;  ( -- n )
;
; Leave the number of the quantity of 16-bit values contained in
; the data stack, before n added.
;
; : DEPTH 
;   S0 @ SP@ - 2/ ;
;
        head DEPTH,5,"DEPTH",docolon
        dw S0,FETCH,SPFETCH
        dw MINUS,TWODIV,EXIT

; ----------------
;  THE SPACE WORD
; ----------------
;  ( -- )
;
; Transmit an ASCII blank to the current output device.
;
; : SPACE   BL EMIT ;
;
        head SPACE,5,"SPACE",docode
        push    bc
        ld      c,$20
        jp      EMIT

; -----------------
;  THE SPACES WORD
; -----------------
;  ( n -- )
;
; Transmit n spaces to the current output device. Take no action
; for n of zero or less.
;
; : SPACES  0 MAX ?DUP IF 0 DO SPACE LOOP THEN ;
;
        head SPACES,6,"SPACES",docode
        MODULE SPACES
        ld      a,b
        or      a
        jp      m,L02           ; finish if n is negative
L01     ld      a,b
        or      c
        jr      z,L02           ; finish if n is null
        ld      a,$20
        call    prt_a
        dec     bc
        jr      L01
L02     pop     bc
        jp      (iy)
        ENDMODULE

; -----------------
;  THE EXPECT WORD
; -----------------
;  ( addr n -- )
;
; Transfer characters from the terminal beginning at addr, upward,
; until a "return" or the count of n has been received. Take no
; action for n less than or equal to zero. One or two nulls are
; added at the end of text.
;
; : EXPECT      OVER + OVER
;               DO 14 EMIT KEY 15 EMIT DUP 8 =
;                  IF DROP DUP I = DUP R> 2- + >R
;                     IF 0 ELSE 8 THEN
;                  ELSE DUP 0D = IF LEAVE DROP BL 0
;                                ELSE DUP
;                                THEN
;                  I C! 0 I 1+ !
;                  THEN EMIT
;               LOOP DROP ;
;
        head EXPECT,6,"EXPECT",docode
        MODULE EXPECT
        pop     hl              ; HL = c-addr
        push    de              ; save IP
        ld      b,0             ; B = received count, C = expected count
L01     ld      a,b             ; received = expected ?
        cp      c
        jr      z,L05
L02     ld      a,$0e           ; cursor on
        call    prt_a           ;
        push    hl
        ld      de,L03
        jp      PKEY            ; wait for a key
L03     dw      L04
L04     ld      a,$0f           ; cursor off
        call    prt_a           ;
        ld      a,c
        pop     bc              ; align stack
        pop     hl
        cp      $20             ; jump if a printable character
        jr      nc,L06          ;
        cp      $08             ; Backspace ?
        jr      z,L08           ;
        cp      $01             ; del_line?
        jr      z,L09           ;
        cp      $0d             ; ENTER ?
        jr      nz,L02          ;

;ENTER ($0D) was pressed
L05     ;ld      a,$0f           ; cursor off
        ;call    prt_a           ;
        ld      a,$20           ; print a space
        call    prt_a           ;
        ld      (hl),0          ; terminate the input string with zero
        pop     de              ; restore IP
        pop     bc              ; new TOS
        jp      (iy)

L06     ld      (hl),a          ; save character in input buffer
        inc     hl              ; increment pointer
        inc     b               ; increment char count
L07     call    prt_a           ; emit char
        jr      L01

L08     xor     a               ; is in start of the input buffer?
        cp      b               ;
        jr      z,L02
        dec     hl              ; decrement pointer
        dec     b               ; decrement char counter
        ld      a,$08
        jr      L07             ;

L09     xor     a               ; is in start of the input buffer?
        cp      b               ;
        jr      z,L02
L10     ld      a,$08           ; backspace
        call    prt_a
        dec     hl              ; decrement pointer
        djnz    L10
        jr      L01             ;
        ENDMODULE

; ----------------
;  THE COUNT WORD
; ----------------
;  ( addr -- addr+1 b )
;
; Leave the address addr+1 and the character count of text beginning
; at addr. The first byte at addr must contain the character count b.
; Range of b is {0..255}.
;
; : COUNT   DUP 1+ SWAP C@ ;
;
        head COUNT,5,"COUNT",docode
        ld      a,(bc)
        inc     bc
        push    bc
        ld      c,a
        ld      b,0
        jp      (iy)

; ---------------
;  THE TYPE WORD
; ---------------
;  ( addr n -- )
;
; Transmit n characters beginning at address to the current output
; device. No action takes place for n less than or equal to zero.
;
; : TYPE
;     DUP 0>
;     IF OVER + SWAP
;        DO I C@ EMIT LOOP
;     ELSE 2DROP
;     THEN ;
;
        head TYPE,4,"TYPE",docode
        MODULE TYPE
        pop     hl                      ; addr to HL, n in BC
        bit     7,b
        jr      nz,L02                  ; done if n negative

L01     ld      a,b
        or      c
        jr      z,L02                   ; done if n = 0
        ld      a,(hl)
        call    prt_a
        inc     hl
        dec     bc
        jr      L01

L02     pop     bc                      ; new TOS
        jp      (iy)
        ENDMODULE

; --------------------
;  THE -TRAILING WORD
; --------------------
;  ( addr n1 -- addr n2 )               "dash-trailing"
;
; Adjust the character count n1 of a text string beginning at addr
; to exclude trailing blanks, i.e., the characters at addr+n2 to
; addr+n1-1 are blanks. An error condition exists if n1 is negative.
;
; : -TRAILING
;   DUP 0
;   DO
;     2DUP + 1- C@ BL -
;     IF LEAVE ELSE 1-
;     THEN
;   LOOP ;
;
        head TRAILING,9,"-TRAILING",docode
        MODULE TRAILING
        pop hl
        push hl
        add hl,bc
L01     dec hl
        ld a,(hl)
        cp $20
        jp nz,NEXT
        dec bc
        jr L01
        ENDMODULE

; ---------------
;  THE HERE WORD
; ---------------
;  ( -- addr )
;
; Return the address of the next available dictionary location.
;
; : HERE   DP @ ;
;
        head HERE,4,"HERE",docode
        push    bc
        ld      bc,(dp)
        jp      (iy)

; ------------
;  THE , WORD
; ------------
;  ( x -- )                     "comma"
;
; Allot two bytes in the dictionary, storing n there.
;
        head COMMA,1,",",docode
        ld      hl,(dp)
        ld      (hl),c
        inc     hl
        ld      (hl),b
        inc     hl
        ld      (dp),hl
        pop     bc
        jp      (iy)

; -------------
;  THE C, WORD
; -------------
;  ( c --   )                   "c-comma"
;
; Store the low-order 8 bits of n at the next byte in  the dictionary,
; advancing the dictionary pointer.
;
        head CCOMMA,2,"C,",docode
        ld      hl,(dp)
        ld      (hl),c
        inc     hl
        ld      (dp),hl
        pop     bc             ;new TOS
        jp      (iy)

; ---------------
;  THE WORD WORD
; ---------------
;  ( c -- addr )
;
; Receive characters from the input stream until the  non-zero
; delimiting character is encountered or the input stream  is
; exhausted, ignoring leading delimiters. The characters are stored
; as a packed string with the character count in the first character
; position. The actual delimiter encountered (char or null) is
; stored at the end of the text but not included in the count.
; If the input stream was exhausted as WORD is called, then a zero
; length will result. The address of the beginning of this packed
; string is left on the stack.
;
; : WORD        BLK @ ?DUP IF BLOCK ELSE TIB THEN
;               SWAP OVER >IN @ + (WORD) SWAP -
;               >IN ! HERE ;
;
        head WORD,4,"WORD",docolon
        dw BLK,FETCH,QDUP,QBRANCH,WRD1
        dw BLOCK,BRANCH,WRD2
WRD1:   dw TIB
WRD2:   dw SWAP,OVER,TOIN,FETCH,PLUS,PWORD
        dw SWAP,MINUS,TOIN,STORE,HERE,EXIT

        nohead PWORD,docode
        MODULE PWORD
        ld h,b                  ; c-addr to HL
        ld l,c
        pop bc                  ; delimiter to C, B=0
        push de                 ; sv IP
L02     ld a,(hl)               ; get a character
        cp c                    ; compare with delimiter
        jr nz,L01
        inc hl
        jr L02                  ; ignore delimiters before text

L01     push hl                 ; save the start string address
L05     ld a,(hl)               ; delimiter char?
        cp c                    ;
        jr z,L03                ;
        or a                    ; null?
        jr z,L04
        inc b                   ; increment counter
        inc hl                  ; increment pointer
        jr L05

L03     inc hl
L04     ld de,(dp)              ; DE = dest address of string
        ld a,b                  ; save string lenght
        ld (de),a               ;
        ex (sp),hl              ; restore the start string address and
                                ; preserve the actual stream address
        or a                    ; null lenght?
        jr z,L06
        ld a,c                  ; save delimiter char
        inc de                  ; copy the string
        ld c,b                  ;
        ld b,0                  ;
        ldir                    ;
        ld (de),a               ; store delimiter at end of text

L06     pop bc                  ; actual stream address
        pop hl                  ; restore IP
        jp      NEXTHL
        ENDMODULE

; ---------------
;  THE FIND WORD
; ---------------
;  ( --- addr )
;
; Leave the compilation address of the next word name,  which is
; accepted from the input stream. If that word cannot be found
; in the dictionary after a search of CONTEXT and FORTH leave
; zero.
;
; : FIND
;   BL WORD CONTEXT @ @ (FIND)     \ Search context vocabulary first.
;   CONTEXT @ CURRENT @ -          \ Is CONTEXT different from CURRENT?
;   IF ?DUP 0= IF                  \ If last find returned 0 flag?
;        HERE CURRENT @ @ (FIND)   \ Search CURRENT.
;      THEN
;   THEN ;
;
        head FIND,4,'FIND',docode
        MODULE FIND
        push    bc
        push    de
        ld      bc,$20
        ld      de,L01
        jp      WORD
L01     dw      L02
L02     ld      d,b             ; address of name to find to DE
        ld      e,c
        ld      hl,(context)
        ld      a,(hl)
        inc     hl
        ld      h,(hl)          ; dictionary word address to HL, points to last
        ld      l,a             ; defined name
        call    L04             ; (FIND)
        ld      hl,(context)
        ld      de,(current)
        or      a
        sbc     hl,de
        jr      z,L03
        ld      a,b
        or      c
        jr      nz,L03
        ld      hl,(dp)         ; address of name to find to DE
        ex      de,hl           ; CURRENT @ to HL
        ld      a,(hl)
        inc     hl
        ld      h,(hl)          ; dictionary word address to HL, points to last
        ld      l,a             ; defined name
        call    L04             ; (FIND)
L03     pop     de
        jp      (iy)

; The runtime (FIND) code
;  ( addr1 addr2 --- addr3 )
;
L04     push    de              ; save address name
        push    hl              ; save dictionary word address
        ld      c,$00
        ld      a,(de)          ; get the word lenght
        xor     (hl)            ; compare with length of the word in dictionary
        and     $3f             ; mask immed bit
        jr      nz,L06          ; if d'nt match, advance to next word in dictionary
        xor     (hl)
        and     $3f
        ld      b,a             ; lenght at B to be a loop counter
L05     inc     hl              ; compare the two words
        inc     de              ;
        ld      a,(de)          ;
        cp      (hl)            ;
        jr      nz,L06          ; if d'nt match, advance to next word in dictionary
        djnz    L05
        inc     hl
        ld      b,h             ; and transfer it to BC
        ld      c,l             ;
        pop     de              ; discard dictionary word address
        pop     de              ; discard c-addr
        ret

L06     pop hl                  ; restore dictionary word address
        dec     hl              ; get the link to next word in dictionary
        ld      d,(hl)          ;
        dec     hl
        ld      e,(hl)          ;
        ex      de,hl           ;
        pop     de              ; restore c-addr
        ld      a,h             ; test for end o dictionary (link = 0)
        or      l               ;
        jr      nz,L04         ; if nonzero, jump to continue the search
        ld      b,l             ; word not found, leave flag = 0 in TOS
        ret
        ENDMODULE

; --------------
; THE DIGIT WORD
; --------------
;  ( c -- n tf )  (ok)
;  ( c -- ff )    (bad)
;
; Converts the ASCII character c to its binary, equivalent n, accompanied
; by a true flag. If the conversion is invalid, leaves only a false flag.
;
        nohead DIGIT,docode
        MODULE DIGIT
        ld a,c                  ; ASCII CHR to be converted
        sub '0'                 ; >=0?
        jr c,L02                ; <0 is invalid
        cp 10                   ; >9?
        jr c,L01                ; no, test BASE value
        sub 7                   ; gap between "9" & "A"
        cp 10                   ; >="A"?
        jr c,L02                ; CHRs between "9" & "A" are invalid

L01     ld hl,(UP)              ; read user pointer to HL
        ld bc,8
        add hl,bc               ; Points to the BASE variable
        cp (hl)                 ; < BASE value?

        jr nc,L02               ; no, invalid
        ld c,a                  ; converted digit
        push bc                 ;
        ld c,1                  ; true flag
        jp      (iy)

L02     ld c,0
        jp      (iy)
        ENDMODULE

; ------------------
;  THE CONVERT WORD
; ------------------
;  ( +d1 addr1 -- +d2 addr2 )                             79
;
; +d2 is the result of converting the characters within the text beginning
; at addr1+1 into digits, using the value of BASE, and accumulating each
; into +d1 after multiplying +d1 by the value of BASE. Conversion continues
; until an unconvertible character is encounter.  addr2 is the location of
; the first unconvertible character.
;
; : CONVERT ( ud1 addr1 --- ud2 addr2)
;   BEGIN
;      1+             \ Increment address.
;      DUP >R         \ Save it on the return stack
;      C@ DIGIT       \ Read a character and convert it to digit value.
;   WHILE
;      SWAP BASE @ U* DROP \ Multiply number MSW by BASE. (digitval msw*base)
;      ROT BASE @ U* D+ \ Multiply number LSW by base and add it.
;                        \ Result is double number with value ud1*BASE+digitval.
;      DPL @ 1+ IF 1 DPL +! THEN \ Increments DPL if it was set earlier..
;      R>             \ Get saved address.
;   REPEAT R> ;
;
        head CONVERT,7,"CONVERT",docolon
CONV1:  dw ONEPLUS,DUP,TOR,CFETCH,DIGIT         ; BEGIN
        dw QBRANCH,CONV2                        ; WHILE
        dw SWAP,BASE,FETCH,UTIMES,DROP
        dw ROT,BASE,FETCH,UTIMES,DPLUS
        dw DPL,FETCH,ONEPLUS,QBRANCH,CONV3      ; IF
        dw ONE,DPL,PLUSSTORE
CONV3:  dw FROMR,BRANCH,CONV1                   ; THEN  REPEAT
CONV2:  dw FROMR,EXIT

; -----------------
;  THE ERRNUM WORD
; -----------------
;  ( f -- )                                   "error-number"
; If NUMBER can't convert the word after the interpreter could
; not find it, then abort.
;
        head ERRNUM,6,"ERRNUM",docode
        ld      hl,PERRNUM
        jp      (hl)

; -----------------
;  THE NUMBER WORD
; -----------------
;  ( addr -- d )
;
; Convert the count and character string at addr, to a signed
; 32-bit integer, using the current base. If numeric conversion
; is not possible, an error condition exists. The string  may
; contain a preceding negative sign.
;
;   0 0                      \ Start with a double of value 0.
;  ROT DUP 1+ C@             \ Read character at addr+1
;      2D = DUP >R           \ Compare character with '-', save to return stack.
;       +                    \ Also use the flag to adjust the address
;                            \ (if there was a minus sign, skip to next character.
;      -1                    \ Initialize DPL to -1.
;      BEGIN
;         DPL !              \ Store DPL value (-1 or 0).
;         CONVERT            \ Convert a string of digits into binary.
;         DUP C@ BL -        \ Is the next character a blank space?
;      WHILE
;         DUP C@ 2E - IF     \ Compare the next character with '.'
;         ERRNUM             \ Error if not equal.
;         THEN  0            \ Initialize DPL to 0.
;      REPEAT
;      DROP                  \ drop address.
;      R> IF DNEGATE THEN ;  \ negate the number if saved sign
;
        head NUMBER,6,"NUMBER",docolon
        dw ZERO,ZERO,ROT,DUP,ONEPLUS,CFETCH,LIT,'-'
        dw EQUAL,DUP,TOR,PLUS,LIT,-1
NUMB1:  dw DPL,STORE,CONVERT,DUP,CFETCH,BL,MINUS        ; BEGIN
        dw QBRANCH,NUMB2,DUP,CFETCH,LIT,'.',MINUS       ; WHILE
        dw QBRANCH,NUMB3,ERRNUM                         ; IF
NUMB3:  dw ZERO,BRANCH,NUMB1                            ; THEN REPEAT
NUMB2:  dw DROP,FROMR,QBRANCH,NUMB4,DNEGATE             ; IF
NUMB4:  dw EXIT                                         ; THEN

; ------------
;  THE [ WORD
; ------------
;  ( -- )                               "left-bracket"
;
; End the compilation mode. The text from the input stream is
; subsequently executed.  See ]
;
; : [   0 STATE ! ; IMMEDIATE
;
        immed LEFTBRKT,1,"[",docode
        xor     a
        ld      (state),a
        jp      (iy)

; ------------
;  THE ] WORD
; ------------
;  ( -- )                               "right-bracket"
;
; Sets the compilation mode. The text from the input stream is
; subsequently compiled.  See [
;
; : ]   C0 STATE ! ;
;
        head RIGHTBRKT,1,"]",docode
        ld      a,$c0
        ld      (state),a
        jp      (iy)

; --------------------
;  THE IMMEDIATE WORD
; --------------------
;  ( -- )
;
; Marks the most recently made dictionary entry as a word which
; will be executed when encountered during compilation rather
; than compiled.
;
; : IMMEDIATE   LATEST DUP C@ 40 OR SWAP C! ;
;
        head IMMEDIATE,9,"IMMEDIATE",docode
        ld hl,(current)
        ld a,(hl)
        inc hl
        ld h,(hl)
        ld l,a
        ld a,(hl)
        or $40
        ld (hl),a
        jp (iy)

; ---------------
;  THE >NFA WORD
; ---------------
;  ( cfa -- nfa )
;
; Converts the code field address to the name field address
;
        head NFA,4,">NFA",docode
        dec     bc
        ld      a,(bc)
        rla
        jr      nc,NFA
        jp      (iy)

; ---------------
;  THE >PFA WORD
; ---------------
;  ( cfa -- pfa )
;
; Converts the code field address to the parameter field address
;
; : >PFA  3 + ;
;
        head PFA,4,">PFA",docode
        inc     bc
        inc     bc
        inc     bc
        jp      (iy)

; ---------------
;  THE >CFA WORD
; ---------------
;  ( nfa -- cfa )
;
; Converts the name field address to the code field address
;
; : >CFA  COUNT 1F AND + ;
;
        head CFA,4,">CFA",docode
        ld      a,(bc)
        inc     bc
        and     $1f
        add     c
        ld      c,a
        ld      a,0
        adc     b
        ld      b,a
        jp      (iy)

; -------------
;  THE BL WORD
; -------------
;  ( -- c )                             "b-l"
;
; Leaves the ASCII character value for space (decimal 32).
;
        head BL,2,"BL",docon
        dw 0020h

; --------------
;  THE TIB WORD
; --------------
;  ( -- addr )                   83             "t-i-b"
;  The address of the text input buffer.  This buffer is used
;  to hold characters when the input stream is coming from the
;  current input device.  The minimum capacity of TIB is 80
;  characters.
;
        head TIB,3,"TIB",docon
tib:    dw 0

; ---------------
;  THE #SCR WORD
; ---------------
;  ( -- n )
;
; Returns number of screens in RAM disk.
;
        head SHARPSCR,4,"#SCR",docon
        dw 8

; ----------------
;  THE B/BUF WORD
; ----------------
;  ( -- 1024 )
;
; Constant, the number of bytes in a block buffer.
;
        head BBUF,5,"B/BUF",docon
        dw 400h

; ----------------
;  THE B/SCR WORD
; ----------------
;  ( -- 1 )
;
; Constant, the number of block buffers per screen.
;
        head BSCR,5,"B/SCR",docon
        dw 01h

; --------------
;  THE C/L WORD
; --------------
;  ( -- 32 )
;
; Returns the number of characters per line.
;
        head CL,3,"C/L",docon
        dw 20h

; ----------------
;  THE FIRST WORD
; ----------------
;  ( -- addr )
;
; Returns the first address of the block buffer.
;
        head FIRST,5,"FIRST",docon
first:  dw 0

; ----------------
;  THE LIMIT WORD
; ----------------
;  ( -- addr )
;
; The address just above the block buffer.
;
        head LIMIT,5,"LIMIT",docon
limit:  dw 0

; -------------
;  THE LO WORD
; -------------
;  ( -- addr )
;
; Variable containing the bottom address of the RAM disk.
;
        head LO,2,"LO",dovar
lo:     dw $8000

; -------------
;  THE S0 WORD
; -------------
;  ( -- addr )                                   "s-zero"
; Returns the address of the bottom of the stack, when empty.
;
        head S0,2,"S0",douser
        db 4

; -------------
;  THE R0 WORD
; -------------
;  ( -- addr )                                   "r-zero"
; Returns the address of the bottom of the return stack, when empty.
;
        head R0,2,"R0",douser
        db 6

; ---------------
;  THE BASE WORD
; ---------------
;  ( -- addr )
;
; Leaves the address of a variable containing the current input-output
; numeric conversion base.  {2..70}
;
        head BASE,4,"BASE",douser
        db 8

; -------------
;  THE DP WORD
; -------------
;  ( -- addr )
;
; Leaves the address of a variable containing the dictionary pointer,
; which addresses the next free memory above the dictionary
;
        head DP,2,"DP",docon
        dw usr_addr+10

; ----------------
;  THE FENCE WORD
; ----------------
;  ( -- addr )
;  Address below which dictionary is not cleared.
;
        head FENCE,5,"FENCE",docon
        dw usr_addr+12

; ----------------
;  THE STATE WORD
; ----------------
;  ( -- addr )
;
; Leaves the address of the variable containing the compilation state.
; A non-zero content indicates compilation is occurring, but the value
; itself may be installation dependent.
;
        head STATE,5,"STATE",docon
        dw usr_addr+14

; --------------
;  THE >IN WORD
; --------------
;  ( -- addr )                          "to-in"
;
; Leaves the address of a variable which contains the present
; character offset within the input stream {{0..1023}}
; See:  WORD  (  ."  FIND
;
        head TOIN,3,">IN",docon
        dw usr_addr+16

; --------------
;  THE HLD WORD
; --------------
;  ( -- addr )
;
; Leaves the address of a variable wich contains the current address
; position within the PAD, where HOLD will insert the next character
;
        head HLD,3,"HLD",docon
        dw usr_addr+18

; -------------------
;  THE VOC-LINK WORD
; -------------------
;  ( -- addr )
;
; Leaves the address of the Terminal Input Buffer
;
        head VOC_LINK,8,"VOC-LINK",docon
        dw usr_addr+20

; ------------------
;  THE CONTEXT WORD
; ------------------
;  ( -- addr )
;
; Leave the address of a variable specifying the vocabulary in which
; dictionary searches are to be made, during interpretation of the
; input stream.
;
        head CONTEXT,7,"CONTEXT",docon
        dw usr_addr+22

; ------------------
;  THE CURRENT WORD
; ------------------
;  ( -- addr )
;
; Leave the address of a variable specifying the vocabulary into which
; new word definitions are to be entered.
;
        head CURRENT,7,"CURRENT",docon
        dw usr_addr+24

; --------------
;  THE DPL WORD
; --------------
;  ( -- addr )                   U               "d-p-l"
;
; A variable containing the number of places after the
; fractional point for input conversion.
;
        head DPL,3,"DPL",docon
        dw usr_addr+26

; --------------
;  THE BLK WORD
; --------------
;  ( -- addr )
;
; Leaves the address of a variable containing the number of the mass
; storage block being interpreted as the input stream. If the content
; is zero, the input stream is taken from the terminal.
;
        head BLK,3,"BLK",docon
        dw usr_addr+28

; --------------
;  THE SCR WORD
; --------------
;  ( -- addr )                   U,217
; Leave  the address of a variable containing the number of  the
; screen most recently listed.
;
        head SCR,3,"SCR",docon
        dw usr_addr+30

; --------------
;  THE MOD WORD
; --------------
;  ( n1 n2 -- n3 )
;
; Divide n1 by n2, leaving the remainder n3, with the same sign
; as n1.
;
; : MOD  /MOD DROP ;
;
        head MOD,3,"MOD",docolon
        dw DIVMOD,DROP,EXIT

; ----------------------
;  THE */MOD WORD
; ----------------
;  ( n1 n2 n3 -- n4 n5 )                "times-divide-mod"
;
; Multiply n1 by n2, divide the result by n3 and leave the remainder
; n4 and quotient n5.  A 32-bit intermediate  product is used as for
; */ .  The remainder has the same sign as n1.
;
; : */MOD    ROT ROT 2DUP XOR >R ABS SWAP ABS U*
;            ROT DUP R@ XOR >R ABS U/MOD
;            R> ?NEGATE SWAP R> ?NEGATE SWAP ;
;
        head TIMESDIVMOD,5,"*/MOD",docolon
        dw ROT,ROT,TWODUP,XOR_,TOR,ABS
        dw SWAP,ABS,UTIMES,ROT,DUP,RFETCH
        dw XOR_,TOR,ABS,UDIVMOD,FROMR,PLUSMINUS
        dw SWAP,FROMR,PLUSMINUS,SWAP,EXIT

; -------------
;  THE */ WORD
; -------------
;  ( n1 n2 n3 -- n4 )                   "times-divide"
;
; Multiply n1 by n2, divide the result by n3 and leave the quotient
; n4.  n4 is rounded toward zero. The product of n1 times n2 is
; maintained as an intermediate 32-bit value for greater precision
; than the otherwise equivalent  sequence:  n1 n2 * n3 /
;
; : */   */MOD UNDER ;
;
        head TIMESDIV,2,"*/",docolon
        dw TIMESDIVMOD,UNDER,EXIT

; ----------------
;  THE M/MOD WORD
; ----------------
;  ( ud1 u2 -- u3 ud4 )                 "u-d-divide-mod"
;
; An unsigned mixed magnitude math operation which leaves a double
; quotient ud4 and remainder u3, from a double dividend ud1 and single
; divisor u2.
;
; : M/MOD   >R 0 R@ U/MOD ROT ROT R> U/MOD ROT ;
;
        head MDIVMOD,5,"M/MOD",docolon
        dw TOR,ZERO,RFETCH,UDIVMOD
        dw ROT,ROT,FROMR,UDIVMOD
        dw ROT,EXIT

; --------------
;  THE MIN WORD
; --------------
;  ( n1 n2 -- n3 )                      "min"
;
; Leave the lesser of two numbers.
;
; : MIN   2DUP > IF SWAP THEN DROP ;
;
        head MIN,3,"MIN",docolon
        dw TWODUP,GREATER
        dw QBRANCH,MIN1,SWAP
MIN1:   dw DROP,EXIT

; --------------
;  THE MAX WORD
; --------------
;  ( n1 n2 -- n3 )                      "max"
;
; Leave the greater of two numbers.
;
; : MAX   2DUP < IF SWAP THEN DROP ;
;
        head MAX,3,"MAX",docolon
        dw TWODUP,LESS
        dw QBRANCH,MAX1,SWAP
MAX1:   dw DROP,EXIT

; ------------------
;  THE COMPILE WORD
; ------------------
;  ( -- )
;
; When a word containing COMPILE executes, the 16-bit value following
; the compilation address of COMPILE is copied (compiled) into the
; dictionary. i.e., COMPILE DUP will copy the compilation address of DUP.
;      COMPILE  [ 0 , ]   will copy zero.
;
; : COMPILE   R> DUP 2+ >R @ , ;
;
        head COMPILE,7,"COMPILE",docolon
        dw FROMR,DUP,TWOPLUS,TOR
        dw FETCH,COMMA,EXIT

; --------------------
;  THE [COMPILE] WORD
; --------------------
;  ( -- )
;
; Used in a colon-definition in form:
;     :  xxx    [COMPILE]   FORTH   ;
; [COMPILE] will force the compilation of an immediate definitions,
; that would otherwise execute during compilation. The above example
; will select the FORTH vocabulary then xxx executes, rather than at
; compile time.
;
; : [COMPILE]   FIND , ; IMMEDIATE
;
        immed CCOMPILE,9,"[COMPILE]",docolon
        dw FIND,COMMA,EXIT

; ------------------
;  THE LITERAL WORD
; ------------------
;  ( x --   )  (compiling)
;  ( x -- x )  (executing)
;
; If compiling, then compile the stack value x as a 16-bit literal,
; which when later executed, will leave x on the stack.
;
; : LITERAL   STATE @ IF COMPILE lit , THEN ; IMMEDIATE
;
        immed LITERAL,7,"LITERAL",docolon
        dw STATE,FETCH,QBRANCH,LITRL1
        dw COMPILE,LIT,COMMA
LITRL1: dw EXIT


; --------------------
;  THE DLITERAL WORD
;  ( xd --    )  (compiling)
;  ( xd -- xd )  (executing)
;
; If compiling, compile a stack double number into a literal. Later
; execution of the definition containing the literal will push it to
; the stack. If executing, the number will remain on the stack.
;
; : DLITERAL   STATE @ IF SWAP LITERAL LITERAL THEN ; IMMEDIATE
;
        immed DLITERAL,8,"DLITERAL",docolon
        dw STATE,FETCH,QBRANCH,DLIT1
        dw SWAP,LITERAL,LITERAL
DLIT1:  dw EXIT

; ---------------
;  THE (.") WORD
; ---------------
;  ( -- )
;
; Run-time code for ."
;
; : (.")   R> COUNT 2DUP + >R TYPE ;
;
        nohead XDQUOTE,docolon
        dw FROMR,COUNT,TWODUP
        dw PLUS,TOR,TYPE
        dw EXIT

; -------------
;  THE ." WORD
; -------------
;  ( -- )                                       I,133     "dot-quote"
;
;      Interpreted or used in a colon definition in the form:
;          ." ccc"
;     Accept the following text from the input stream, terminated by
;     "  (double-quote).   If executing,  transmit this text to  the
;     selected output device.   If compiling,  compile so that later
;     execution  will  transmit  the  text to  the  selected  output
;     device.   At least 127 characters are allowed in the text.  If
;     the  input stream is exhausted before the terminating  double-
;     quote, an error condition exists.
;
; : ." 22 STATE @ IF COMPILE (.") WORD C@ 1+ ALLOT
;                  ELSE WORD COUNT TYPE THEN ;
;
        immed DOTQUOTE,2,'."',docolon
        dw LIT,$22,STATE,FETCH,QBRANCH,DQUT1
        dw COMPILE,XDQUOTE,WORD,CFETCH,ONEPLUS
        dw ALLOT,BRANCH,DQUT2
DQUT1:  dw WORD,COUNT,TYPE
DQUT2:  dw EXIT

; -----------------
;  THE LATEST WORD
; -----------------
;  ( -- addr )
;
; Leave the name address of the most recently defined word in CURRENT vocabulary
;
; : LATEST   CURRENT @ @ ;
;
        head LATEST,6,"LATEST",docode
        push    bc
        ld      hl,(current)
        ld      c,(hl)
        inc     hl
        ld      b,(hl)
        jp      (iy)

; --------------
;  THE PAD WORD
; --------------
;  ( -- addr )
;
; The address of a scratch area used to hold character strings for
; intermediate processing. The minimum capacity of PAD is 64 characters
;
; : PAD   HERE 38 + ;
;
        head PAD,3,"PAD",docode
        push    bc
        ld      hl,(dp)
        ld      bc,$38
        add     hl,bc
        ld      b,h
        ld      c,l
        jp      (iy)

; ---------------
;  THE S->D WORD
; ---------------
;  ( n -- d )                           "s-to-d"
;
; Convert a single number n to double number d
;
        head STOD,4,"S->D",docode
        MODULE STOD
        push    bc
        xor     a
        rl      b
        sbc     a,a
        ld      c,a
        ld      b,c
L01     jp      (iy)            ;  ( n1--d1L d1H)
        ENDMODULE

; ---------------
;  THE DABS WORD
; ---------------
;  ( d1 -- d2 )                         "d-absolute"
;
; Leave as a positive double number d2, the absolute value of a
; double number, d1.  {0..2,147,483,647}
;
; DABS   DUP D+- ;
;
        head DABS,4,"DABS",docode
        push    bc
        jp      DPLUSMINUS

; ----------------
;  THE QUERY WORD
; ----------------
;  ( -- )
;
; Accept input of up to 80 characters (or until a 'return') from the
; operator's terminal, into the terminal input buffer. WORD may be used
; to accept text from this buffer as the input stream, by setting >IN
; and BLK to zero.
;
; (OBS.: In this definition, QUERY accepts up to 128 characters)
;
; : QUERY  TIB 80 EXPECT 0 >IN ! 0 BLK ! ;
;
        head QUERY,5,"QUERY",docolon
        dw TIB,LIT,$80,EXPECT,ZERO
        dw TOIN,STORE,ZERO,BLK,STORE,EXIT

; ---------------
;  THE HOLD WORD
; ---------------
;  ( c -- )
;
; Insert char into a pictured numeric output string. May only be used
; between <# and #> .
;
; : HOLD   -1 HLD +! HLD @ C! ;
;
        head HOLD,4,"HOLD",docode
        ld      hl,(hld)
        dec     hl
        ld      (hld),hl
        ld      (hl),c
        pop     bc
        jp      (iy)

; ---------------
;  THE SIGN WORD
; ---------------
;  ( n -- )
;
; Insert the ASCII "-" (minus sign) into the pictured numeric output
; string, if n is negative.
;
; : SIGN   ROT 0< IF ASCII - HOLD THEN ;
;
        head SIGN,4,"SIGN",docode
        pop hl
        ex (sp),hl
        sla h
        jp nc,NEXT
        push bc
        ld c,'-'
        jp HOLD

; -------------
;  THE <# WORD
; -------------
;  ( -- )                               "less-sharp"
;
; Initialize pictured numeric output.  The words:
;      #  #>  #S  <#  HOLD  SIGN
; can be used to specify the conversion of a double-precision number into
; an ASCII character string stored in right-to-left order.
;
; : <#   PAD HLD ! ;
;
        head LESSHARP,2,"<#",docolon
        dw PAD,HLD,STORE,EXIT

; ------------
;  THE # WORD
; ------------
;  ( ud1 -- ud2 )                       "sharp"
;
; Generate from an unsigned double number d1, the next ASCII character
; which is placed in an output string. Result d2 is the quotient after
; division by BASE is maintained for further processing.
; Used between <# and #> .
;
; : #   BASE @ M/MOD ROT 9 OVER < IF 7 + THEN 30 + HOLD ;
;
        head SHARP,1,"#",docolon
        dw BASE,FETCH,MDIVMOD,ROT
        dw TODIGIT,HOLD,EXIT

TODIGIT ld      a,c             ; digit in A
        add     a,$30           ; convert to ASCII
        cp      $3a             ; test for '9'
        jr      c,TD1           ; jump if a number
        add     a,$07           ; else, it is a letter
TD1     ld      c,a             ;
        next

; -------------
;  THE #S WORD
; -------------
;  ( ud -- 0 0 )                        "sharp-s"
;
; Convert all digits of an unsigned 32-bit number ud, adding each to the
; pictured numeric output text, until remainder is zero. A single zero is
; added to the output string if the number was initially zero. Use only
; between <# and #>.
;
; : #S   BEGIN # 2DUP OR 0= UNTIL ;
;
        head SHARPS,2,"#S",docolon
SHRPS1: dw SHARP,TWODUP,OR_         ;BEGIN
        dw ZEQUAL,QBRANCH,SHRPS1    ;UNTIL
        dw EXIT

; -------------
;  THE #> WORD
; -------------
;  ( d -- addr n )                      "sharp-greater"
;
; End pictured numeric output conversion. Drop d, leaving the text ; address, and character count, suitable for TYPE.
;
; : #>   2DROP HLD @ PAD OVER - ;
;
        head SHARPGT,2,"#>",docolon
        dw TWODROP,HLD,FETCH,PAD
        dw OVER,MINUS,EXIT

; --------------
;  THE D.R WORD
; --------------
;  ( d c -- )                           "d-dot-r"
;
; Display d converted according to BASE, right aligned in an c character
; field. Display the sign only if negative.
;
; : D.R  >R SWAP OVER DABS <# #S SIGN #> R>
;        OVER - SPACES TYPE ;
;
        head DDOTR,3,"D.R",docolon
        dw TOR,SWAP,OVER,DABS,LESSHARP
        dw SHARPS,SIGN,SHARPGT,FROMR
        dw OVER,MINUS,SPACES,TYPE,EXIT

; -------------
;  THE .R WORD
; -------------
;  ( n1 n2 -- )                         "dot-r"
;
; Print n1 right aligned in a field of n2 characters, according to BASE.
; If n2 is less than 1, no leading blanks are supplied.
;
; : .R   SWAP S->D ROT D.R ;
;
        head DOTR,2,".R",docode
        pop     hl
        push    hl
        xor     a
        rl      h
        sbc     a,a
        ld      h,a
        ld      l,a
        push    hl
        jr      DDOTR

; -------------
;  THE D. WORD
; -------------
;  ( d -- )                             "d-dot"
;
; Display d converted according to BASE in a free field format, with one
; trailing blank. Display the sign only if negative.
;
; : D.   SWAP OVER DABS <# #S SIGN #> TYPE SPACE ;
;
        head DDOT,2,"D.",docolon
        dw SWAP,OVER,DABS,LESSHARP
        dw SHARPS,SIGN,SHARPGT,TYPE
        dw SPACE,EXIT

; ------------
;  THE . WORD
; ------------
;  ( n -- )                             "dot"
;
; Display n converted according to BASE in a free field format with one
; trailing blank. Display only a negative sign
;
; : .   S->D D. ;
;
        head DOT,1,".",docolon
        dw STOD,DDOT,EXIT

; -------------
;  THE U. WORD
; -------------
;  ( un -- )                            "u-dot'
;
; Display un converted according to BASE as an unsigned number, in a
; free-field format, with one trailing blank.
;
; : U.   0 D. ;
;
        head UDOT,2,"U.",docode
        push    bc
        ld      bc,0
        jr      DDOT

; ------------
;  THE ? WORD
; ------------
;  ( addr -- )                   194    "question-mark"
;
; Display the number at address, using the format of "." .
;
; : ?   @ . ;
;
        head QUEST,1,"?",docode
        ld      h,b
        ld      l,c
        ld      c,(hl)
        inc     hl
        ld      b,(hl)
        jr      DOT

; ------------------
;  THE DECIMAL WORD
; ------------------
;  ( -- )
;
; Set the input-output numeric conversion base to ten.
;
; : DECIMAL   0A BASE ! ;
;
        head DECIMAL,7,"DECIMAL",docode
        push    bc
        ld      bc,10
setbase:
        call    docolon
        dw BASE,STORE,EXIT

; --------------
;  THE HEX WORD
; --------------
;  ( -- )
;
; Set the input-output numeric conversion base to sixteen.
;
; : HEX   10 BASE ! ;
;
        head HEX,3,"HEX",docode
        push    bc
        ld      bc,16
        jr      setbase

; -------------
;  THE H. WORD
; -------------
;  ( u -- )
;
; Prints u as four hexadecimal digit.
;
; : H.  BASE @ SWAP HEX U. BASE ! ;
;
        head HDOT,2,'H.',docolon
        dw BASE,FETCH,SWAP,HEX
        dw UDOT,BASE,STORE,EXIT

; ----------------------------------
; Check that we are not below FENCE.
;  ( addr -- addr )
;
        nohead QFENCE,docolon
        dw DUP,FENCE,FETCH,ULESS,XABORTQ
        db 20,"PROTECTED DICTIONARY"
        dw EXIT

; ----------------
;  THE ALLOT WORD
; ----------------
;  ( n -- )
;
; Add n bytes to the parameter field of the most recently defined word.
;
; : ALLOT
;   HERE +                        \ Add to current dictionary pointer.
;   SP@ 1- OVER 0C + U<           \ Check that we have enough space below SP.
;   ABORT" Dictionary full"
;   ?fence STOPOFF DP !  STOPON ; \ Prevent break key interrupt while updating DP.
;
        head ALLOT,5,"ALLOT",docolon
        dw HERE,PLUS,S0,FETCH,LIT,$10C,MINUS
        dw OVER,ULESS,XABORTQ
        db 15,"DICTIONARY FULL"
        dw QFENCE,STOPOFF,DP,STORE,STOPON,EXIT

; ---------------------------------------
; ?STACK   ( -- )    test for empty stack
;    S0 @ SP@ 2+ U<
;    ABORT" EMPTY STACK" ;
;
        nohead QSTACK,docolon
        dw S0,FETCH,SPFETCH,TWOPLUS,ULESS,XABORTQ
        db 11,"EMPTY STACK"
        dw EXIT

; --------------------
;  THE INTERPRET WORD
; --------------------
;  ( -- )
;
; The outer text interpreter which sequentially executes or compiles text
; from the input stream (terminal or disc) depending on STATE. If the word
; name cannot be found after a search of CONTEXT and then CURRENT it is
; converted to a number according to the current base.
; That also failing, an error message echoing the name with a " ?" will be
; given. Text input will be taken according to the convention for WORD. If
; a decimal point is found as part of a number, a double number value will
; be left. The decimal point has no other purpose than to force this
; action. See NUMBER.
;
; : INTERPRET
;   BEGIN
;      FIND HERE C@
;   WHILE
;      ?DUP
;      IF DUP >NFA C@ STATE @ <
;         IF , ELSE EXECUTE THEN
;      ELSE HERE NUMBER DPL @ 1+
;           IF DLITERAL
;           ELSE DROP LITERAL THEN
;      THEN
;      ?STACK
;   REPEAT
;   DROP ;
;
        head INTERPRET,9,"INTERPRET",docolon
INTE1:  dw FIND                 ;BEGIN
        dw HERE
        dw CFETCH
        dw QBRANCH,INTE7        ;WHILE
        dw QDUP
        dw QBRANCH,INTE2        ;IF
        dw DUP
        dw NFA
        dw CFETCH
        dw STATE
        dw FETCH
        dw LESS
        dw QBRANCH,INTE3        ;IF
        dw COMMA
        dw BRANCH,INTE5         ;ELSE
INTE3:  dw EXECUTE              ;THEN
INTE4:  dw BRANCH,INTE5         ;ELSE
INTE2:  dw HERE
        dw NUMBER
        dw DPL
        dw FETCH
        dw ONEPLUS
        dw QBRANCH,INTE6        ;IF
        dw DLITERAL
        dw BRANCH,INTE5         ;ELSE
INTE6:  dw DROP
        dw LITERAL              ;THEN
INTE5:  dw QSTACK               ;THEN
        dw BRANCH,INTE1         ;REPEAT
INTE7:  dw DROP
        dw EXIT

; ---------------
;  THE QUIT WORD
; ---------------
;  ( -- )
;
; Clear the return stack, setting execution mode, and return control to
; the terminal.  No message is given.
;
; : QUIT
;    RP! [COMPILE] [
;    BEGIN
;      CR QUERY INTERPRET
;      STATE @ 0= IF ." OK" THEN
;    AGAIN ;
;
        head QUIT,4,"QUIT",docolon
        dw RPSTORE,LEFTBRKT
QLOOP:  dw CR,QUERY,INTERPRET   ; BEGIN
        dw STATE,FETCH,ZEQUAL
        dw QBRANCH,QUIT1,XDQUOTE
        db 3," OK"
QUIT1:  dw BRANCH,QLOOP         ; AGAIN

; --------------
;  THE (') WORD
; --------------
;  ( -- addr )
;
; Reads word from input stream, looks it up in the dictionary and returns
; the compilation address, error message if the word is not found.
;
; : (')   FIND DUP 0= ABORT" ?" ;
;
        head PTICK,3,"(')",docolon
        dw FIND,DUP,ZEQUAL,PERRNUM
        dw EXIT

; ------------
;  THE ' WORD
; ------------
;  ( -- addr )          Modified 79             "tick"
;
; Used in the form:
;       '  <name>
; If executing, leave the code field address of the next word accepted
; from the input stream. If compiling, compile this address as a literal;
; later execution will place this value on the stack. An error condition
; exists if not found after a search of the CONTEXT and FORTH vocabularies.
; Within a colon-definition ' <name> is identical to [ ' <name> ] LITERAL.
;
; : '   (') [COMPILE] LITERAL ; IMMEDIATE
;
        immed TICK,1,"'",docolon
        dw PTICK,LITERAL,EXIT

; ----------------
;  THE ASCII WORD
; ----------------
;  ( -- c )
;
; Leave the ASCII character value of the next non-blank character in the
; input stream. If compiling, compile it as a literal, which will be later
; left when executed.
;
; : ASCII   BL WORD 1+ C@ LITERAL ; IMMEDIATE
;
        immed ASCII,5,"ASCII",docolon
        dw BL,WORD,ONEPLUS
        dw CFETCH,LITERAL,EXIT

; ----------------
;  THE !CF   WORD
; ----------------
;  ( adrs cfa -- )
;
; Set code action of a word
;
; : !CF   CD OVER C!        store 'CALL adrs' instr
;         1+ ! ;
;
        nohead STORECF,docolon
        dw LIT,$cd,OVER,CSTORE
        dw ONEPLUS,STORE,EXIT

; ----------------
;  THE ,CF   WORD
; ----------------
;  ( adrs -- )
;
; append a code field
;
; : ,CF   HERE !CF 3 ALLOT ;
;
        nohead COMMACF,docolon
        dw HERE,STORECF,THREE,ALLOT,EXIT

; -----------------
;  THE SMUDGE WORD
; -----------------
;  ( -- )
;
; Toggle bit 5 in name field of latest def to make it findable or unfindable.
;
; : SMUDGE LATEST 20 TOGGLE ;
;
        head SMUDGE,6,"SMUDGE",docolon
        dw LATEST,LIT,$20,TOGGLE,EXIT

; -----------------
;  THE CREATE WORD
; -----------------
;  ( -- )
;
; A defining word used in the form:
;      CREATE  <name>
; to create a dictionary entry for <name>, without allocating any
; parameter field memory. When <name> is subsequently executed, the
; address of the first byte of <name>'s parameter field is left on the
; stack.
;
; : CREATE
;    LATEST ,                  link field
;    BL WORD DUP C@ OVER
;    80 TOGGLE 1+ ALLOT        name field
;    STOPFF                    no BREAK key while updating current
;    CURRENT @ !               new "latest" link
;    docreate ,CF ;            ,CF will call ALLOT will call STOPON
;
        head CREATE,6,"CREATE",docolon
        dw LATEST,COMMA,BL,WORD,DUP,CFETCH
        dw OVER,LIT,$80,TOGGLE,ONEPLUS
        dw ALLOT,STOPOFF,CURRENT,FETCH,STORE
        dw LIT,docreate,COMMACF,EXIT

; ------------------
;  THE (;CODE) WORD
; ------------------
;  ( -- )
;
; The run-time action of DOES>
;
; : (;CODE)   R>        adrs of headless DOES> def'n
;    LATEST >CFA        code field to fix up
;    !CF ;
;
        head XSCCODE,7,"(;CODE)",docolon
        dw FROMR,LATEST,CFA
        dw STORECF,EXIT

; ----------------
;  THE DOES> WORD
; ----------------
;  ( -- )                               "does"
;
; Define the run-time action of a word created by a high-level defining
; word. Used in the form:
;     :  <name> ... CREATE ... DOES> ... ;
;     and then <namex> <name>
; Marks the termination of the defining part of the defining word <name>
; and begins the defining of the run-time action for words that will later
; be defined by <name>. On execution of <namex> the sequence of words
; between DOES> and ; are executed, with the address of <namex>'s
; Parameter field on the stack.
;
; : DOES>   COMPILE (;CODE) dodoes
;           ,CF ; IMMEDIATE
;
        immed DOES,5,"DOES>",docolon
        dw COMPILE,XSCCODE,LIT,dodoes
        dw COMMACF,EXIT

; ------------
;  THE : WORD
; ------------
;  ( -- )                               "colon"
;
; A defining word executed in the form:
;      :  <name>  ...  ;
; Select the CONTEXT vocabulary to be identical to CURRENT.
; Create a dictionary entry for <name> in CURRENT, and set compile mode.
; Words thus defined are called 'colon-definitions'. The compilation
; addresses of subsequent words from the input stream which are not
; immediate words are stored in the dictionary to be executed when <name>
; is later executed. IMMEDIATE words are executed as encountered.
;
; If a word is not found after a search of the CONTEXT and FORTH
; vocabularies, conversion and compilation of a literal number is
; attempted, with regard to the current BASE; that failing, an error
; condition exists.
;
; : :   CURRENT @ CONTEXT ! CREATE
;       SMUDGE docolon HERE 2- ! ]  ;
;
        head COLON,1,":",docolon
        dw CURRENT,FETCH,CONTEXT,STORE,CREATE
        dw SMUDGE,LIT,docolon,HERE,TWOMINUS
        dw STORE,RIGHTBRKT,EXIT

; ------------
;  THE ; WORD
; ------------
;  ( -- )                               "semi-colon"
;
; Terminate a colon definition and stop compilation. If compiling from
; mass storage and the input stream is exhausted before encountering ;
; an error condition exists.
;
; : ;   SMUDGE COMPILE EXIT [ ; IMMEDIATE
;
        immed SEMICOLON,1,";",docolon
        dw SMUDGE,COMPILE,EXIT,LEFTBRKT,EXIT

; -------------
;  THE IF WORD
; -------------
;  ( f -- )
;
; Used in a colon-definition in the form:
;      flag  IF ... ELSE ... THEN   or
;      flag  IF ... THEN
; If flag is true, the words following IF are executed and the words
; following ELSE are skipped. The ELSE part is optional.
; If flag is false, words between IF and ELSE, or between IF and THEN
; (when no ELSE is used), are skipped. IF-ELSE-THEN conditionals may be
; nested.
;
; : IF   COMPILE ?branch HERE 0 , ; IMMEDIATE
;
        immed IF,2,"IF",docolon
        dw COMPILE
        dw QBRANCH
        dw HERE
        dw ZERO
        dw COMMA
        dw EXIT


; ---------------
;  THE THEN WORD
; ---------------
;  ( -- )
;
; Used in a colon-definition in the form:
;      IF ... ELSE ... THEN   or
;      IF ... THEN
; THEN is the point where execution resumes after ELSE or IF (when no
; ELSE is present).
;
; : THEN  HERE SWAP ! ; IMMEDIATE
;
        immed THEN,4,"THEN",docolon
        dw HERE
        dw SWAP
        dw STORE
        dw EXIT

; ---------------
;  THE ELSE WORD
; ---------------
;  ( -- )
;
; Used in a colon-definition in the form:
;      IF ... ELSE ... THEN
; ELSE executes after the true part following IF. ELSE forces execution
; to skip till just after THEN. It has no effect on the stack. (see IF)
;
; : ELSE   COMPILE branch HERE 0 ,
;          SWAP [COMPILE] THEN ; IMMEDIATE
;
        immed ELSE,4,"ELSE",docolon
        dw COMPILE
        dw BRANCH
        dw HERE
        dw ZERO
        dw COMMA
        dw SWAP
        dw THEN
        dw EXIT

; ----------------
;  THE BEGIN WORD
; ----------------
; ( -- )
;
; Used in a colon-definition in the form:
;      BEGIN ... flag UNTIL   or
;      BEGIN ... flag WHILE ... REPEAT
; BEGIN marks the start of a word sequence for repetitive execution.  A
; BEGIN-UNTIL loop will be repeated until flag is true.
; A BEGIN-WHILE-REPEAT loop will be repeated until flag is false. The
; words after UNTIL or REPEAT will be executed when eitherloop is
; finished. flag is always dropped after being tested.
;
; : BEGIN   HERE ; IMMEDIATE
;
        immed BEGIN,5,"BEGIN",docode
        jp      HERE

; ----------------
;  THE UNTIL WORD
; ----------------
;  ( f -- )
;
; Within a colon-definition, mark the end of a BEGIN-UNTIL loop, which
; will terminate based on flag. If flag is true, the loop is terminated.
; If flag is false, execution returns to the first word after BEGIN.
; BEGIN-UNTIL structures may be nested.
;
; : UNTIL   COMPILE ?branch , ; IMMEDIATE
;
        immed UNTIL,5,"UNTIL",docolon
        dw COMPILE
        dw QBRANCH
        dw COMMA
        dw EXIT

; ----------------
;  THE AGAIN WORD
; ----------------
;  ( -- )
;
; Effect an unconditional jump back to the start of a BEGIN-AGAIN loop.
;
; : AGAIN   COMPILE BRANCH , ; IMMEDIATE
;
        immed AGAIN,5,"AGAIN",docolon
        dw COMPILE
        dw BRANCH
        dw COMMA
        dw EXIT

; ------------------
;  THE WHILE REPEAT
; ------------------
;  ( f -- )
;
; Used in the form:
;      BEGIN ... flag WHILE ... REPEAT
; Select conditional execution based on flag. On a true flag, continue
; execution through to REPEAT, which then returns back to just after
; BEGIN. On a false flag, skip execution to just after REPEAT, exiting
; the structure.
;
; : WHILE   COMPILE ?branch HERE DUP , ; IMMEDIATE
;
        immed WHILE,5,"WHILE",docode
        jp      IF

; -----------------
;  THE REPEAT WORD
; -----------------
;  ( -- )
;
; Used in a colon-definition in the form:
;      BEGIN ... WHILE ... REPEAT
; At run-time, REPEAT returns to just after the corresponding BEGIN.
;
; : REPEAT   SWAP [COMPILE] AGAIN [COMPILE] THEN ; IMMEDIATE
;
        immed REPEAT,6,"REPEAT",docolon
        dw SWAP
        dw AGAIN
        dw THEN
        dw EXIT

; -------------
;  THE DO WORD
; -------------
;  ( n1 n2 -- )
;
; Used in a colon-definition:
;      DO ... LOOP   or
;      DO ... +LOOP
; Begin a loop which will terminate based on control parameters.
; The loop index begins at n2, and terminates based on the limit n1.
; At LOOP or +LOOP, the index is modified by a positive or negative
; value. The range of a DO-LOOP is determined by the terminating word.
; DO-LOOP may be nested. Capacity for three levels of nesting is
; specified as a minimum for standard systems.
;
; : DO   COMPILE (do) HERE ; IMMEDIATE
;
        immed DO,2,"DO",docolon
        dw COMPILE
        dw XDO
        dw HERE
        dw EXIT

; ---------------
;  THE LOOP WORD
; ---------------
; ( -- )
;
; Increment the DO-LOOP index by one, terminating the loop if the new
; index is equal to or greater than the limit. The limit and index are
; signed numbers in the range {-32,768..32,767}.
;
; : LOOP   COMPILE (loop) , ; IMMEDIATE
;
        immed LOOP,4,"LOOP",docolon
        dw COMPILE
        dw XLOOP
        dw COMMA
        dw EXIT

; ----------------
;  THE +LOOP WORD
; ----------------
;
; Add the signed  increment n to the loop index using the convention
; for +, and compare the total to the limit. Return execution to the
; corresponding DO until the new index is equal to or greater than the
; limit (n>0), or until the new index is less  than the limit (n<0).
; Upon the exiting from the loop, discard the loop control parameters,
; continuing execution ahead. Index and limit are signed integers in the
; range {-32,768..32,767}.
;
; (Comment: It is a historical precedent that the limit for n<0 is
; irregular. Further consideration of the characteristic is unlikely.)
;
; : +LOOP   ( n -- )
;   COMPILE (+loop) , ; IMMEDIATE
;
        immed PLOOP,5,"+LOOP",docolon
        dw COMPILE
        dw XPLOOP
        dw COMMA
        dw EXIT

; -----------------
; : [!]  ( n addr -- )
;   STATE @ IF [COMPILE] LITERAL COMPILE !
;   ELSE ! THEN ;
;
        nohead BRKTSTORE,docolon
        dw STATE,FETCH,QBRANCH,BST1
        dw LITERAL,LIT,STORE,COMMA,EXIT
BST1:   dw STORE,EXIT

; ----------------
;  THE EXEC: WORD
; ----------------
;  ( -- )
;
; Define a deferred word, used in the form:
;    EXEC: <name>
;
; : EXEC: CREATE -3 ALLOT 21 C, ' NOOP , E9 C, ;
;
        head EXECC,5,"EXEC:",docolon
        dw CREATE,LIT,-3,ALLOT
        dw LIT,$21,CCOMMA,LIT,NOOP      ; compile:   ld hl,NOOP
        dw COMMA,LIT,$E9,CCOMMA,EXIT    ;            jp (hl)

; -------------
;  THE IS WORD
; -------------
;  ( -- )
;
; Set a deferred word, used in the form:
;    ' <name1> IS <name2>
; The action of word <name1> is assigned to the deferred word 0=<name2>.
;
; : IS  (') 1+ [!] ; IMMEDIATE
;
        immed IS,2,"IS",docolon
        dw PTICK,ONEPLUS,BRKTSTORE,EXIT

; ----------------
;  THE VALUE WORD
; ----------------
;  ( -- )
;
; Create a value.
;
; : VALUE  CREATE 0 , DOES> (codefragment)
;
        head VALUE,5,"VALUE",docode
        push    bc
        ld      bc,0
        jp      CONSTANT

; -------------
;  THE TO WORD
; -------------
;  ( x -- )
;
; Set a value.
;
; : TO  (') >PFA [!] ; IMMEDIATE
;
        immed TO,2,"TO",docolon
        dw PTICK,PFA,BRKTSTORE,EXIT

; --------------
;  THE +TO WORD
; --------------
;  ( x -- )
;
; Set a value.
;
; : +TO (') >PFA STATE @ IF [COMPILE] LITERAL COMPILE +!
;   ELSE +! THEN ; IMMEDIATE
;
        immed PLUSTO,3,"+TO",docolon
        dw PTICK,PFA,STATE,FETCH
        dw QBRANCH,PTO1,LITERAL,COMPILE,PLUSSTORE
        dw EXIT
PTO1    dw PLUSSTORE,EXIT

; ---------------
;  THE ADDR WORD
; ---------------
;  ( n1 -- addr)
;
; Returns the address of screen n in the RAM disk.
;
; : ADDR
;   1-                                    \ Screen numbers start at 1.
;   DUP #SCR U< 0= ABORT" Out of ramdisk" \ Check this is a valid screen number.
;   B/BUF * LO @ +                        \ Add to address in LO variable.
; ;
        head ADDR,4,"ADDR",docolon
        dw ONEMINUS,DUP,SHARPSCR,ULESS
        dw ZEQUAL,XABORTQ
        db 14,"OUT OF RAMDISK"
        dw BBUF,TIMES,LO,FETCH
        dw PLUS,EXIT

; -----------------------
;  THE SAVE-BUFFERS WORD
; -----------------------
;  ( -- )
;
; Copies the contents of the block buffer to the RAM disk if it contains a valid screen.
;
; : SAVE-BUFFERS          \ Will always copy the buffer to RAM disk, whether
;                         \ it was updated or not.
;   FIRST 2- @ ?DUP       \ Do we have a buffer?
;   IF
;     ADDR FIRST SWAP B/BUF CMOVE \ Copy to RAM disk.
;   THEN ;
;
        head SAVEBUF,12,"SAVE-BUFFERS",docolon
        dw FIRST,TWOMINUS,FETCH,QDUP
        dw QBRANCH,SAVB1,ADDR,FIRST,SWAP
        dw BBUF,CMOVE
SAVB1:  dw EXIT

; ------------------------
;  THE EMPTY-BUFFERS WORD
; ------------------------
;  ( -- )
;
; Marks the block buffer as empty. FLUSH will not copy it to the RAM disk.
;
; : EMPTY-BUFFERS       \ Mark buffer as empty.
;   0 FIRST 2- ! ;
;
        head EMPTYBUF,13,"EMPTY-BUFFERS",docode
        xor     a
        ld      hl,(first)
        dec     hl
        ld      (hl),a
        dec     hl
        ld      (hl),a
        jp      (iy)

; -----------------
;  THE UPDATE WORD
; -----------------
;  ( -- )
;
; Indicates that the contents of the block buffer have been modified and hence must be
; written back with SAVE-BUFFERS. As this system has a fast RAM disk, buffers will
; always be written back.
;
; : UPDATE ;            \ Do nothing, SAVE-BUFFERS will always save.
;
        head UPDATE,6,"UPDATE",docode
        jp      (iy)

; -----------------
;  THE BUFFER WORD
; -----------------
;  ( n -- addr)
;
; Creates an empty block buffer for screen n (without reading it from the RAM disk) and
; returns its address.
;
; : BUFFER 
;   SAVE-BUFFERS
;   FIRST 2- !             \ Mark block buffer number.
;   FIRST B/BUF BLANKS     \ Fill buffer with spaces.
;   FIRST ;
;
        head BUFFER,6,"BUFFER",docolon
        dw SAVEBUF,FIRST,TWOMINUS,STORE
        dw FIRST,BBUF,BLANKS
        dw FIRST,EXIT

; ----------------
;  THE FLUSH WORD
; ----------------
;  ( -- )
;
; Saves the contents of the block buffer to the RAM disk if it is not empty.
; Next it marks the block buffer as empty.
;
; : FLUSH  SAVE-BUFFERS EMPTY-BUFFERS ;
;
        head FLUSH,5,"FLUSH",docolon
        dw SAVEBUF,EMPTYBUF,EXIT

; ----------------
;  THE BLOCK WORD
; ----------------
;  ( n -- addr)
;
; Causes screen n to be loaded into the block buffer and returns the address of that
; block buffer. Any screen contained in the block buffer that has been changed, will be
; copied back to the RAM disk first.
;
; : BLOCK
;   DUP FIRST 2- @ -  \ Does the current block buffer contain a different block
;                     \ from the one requested?
;   IF
;     SAVE-BUFFERS    \ Save buffer to RAM disk.
;     DUP ADDR FIRST B/BUF CMOVE \ Copy data from RAM disk to buffer.
;     FIRST 2- !      \ Mark the block number stored.
;   ELSE
;      DROP           \ Drop block number.
;   THEN
;   FIRST             \ Return the start address of the block buffer.
;   0 LIMIT C! ;      \ Set byte at LIMIT to 0.
;
        head BLOCK,5,"BLOCK",docolon
        dw DUP,FIRST,TWOMINUS,FETCH,MINUS
        dw QBRANCH,BLK1,SAVEBUF,DUP,ADDR,FIRST
        dw BBUF,CMOVE,FIRST,TWOMINUS,STORE
        dw BRANCH,BLK2
BLK1:   dw DROP
BLK2:   dw FIRST,ZERO,LIMIT,CSTORE,EXIT

; -----------------
;  THE DELETE WORD
; -----------------
;  ( -- )
;
; Delete the file with given name.
;
; : DELETE   BL WORD (DELETE) ;
;
        head DELETE,6,"DELETE",docolon
        MODULE DELETE
        dw BL,WORD,L01
L01     push    de
        call    fname2zx
        ld      a,$7f
        call    api_fileop
        call    response40
        pop     de
        pop     bc
        jp      EXIT
        ENDMODULE

; --------------
;  THE PUT WORD
; --------------
;  ( n1 n2 --- )
;
; Reads a name from the input stream and writes the screens n1 through n2 to the mass
; storage medium in a file with that name.
;
; : PUT
;   FLUSH               \ Synchronize RAM disk with buffer.
;   OVER - 1+ B/BUF *   \ Compute file length.
;   SWAP ADDR SWAP  \ Compute base address ( addr size ---)
;   BSAVE   ;
;
        head PUT,3,"PUT",docolon
        dw FLUSH,OVER,MINUS,ONEPLUS,BBUF,TIMES
        dw SWAP,ADDR,SWAP,BSAVE,EXIT

; --------------
;  THE GET WORD
; --------------
;  ( n --- )
;
; Reads a name from the input stream and reads the file with that name from mass storage
; into the RAM disk at screen n and any following screens.
;
; : GET
;  FLUSH               \ Make sure block buffer is empty.
;  ADDR BLOAD ;
;
        head GET,3,"GET",docolon
        dw FLUSH,ADDR,BLOAD,EXIT

; -----------------
;  THE FORMAT WORD
; -----------------
;  ( -- )
;
; Fills the entire RAM disk with blank spaces.
;
; : FORMAT      \ Clear all buffers in RAM disk.
;   #SCR 1+ 1 DO
;      I BUFFER DROP
;   LOOP ;
        head FORMAT,6,"FORMAT",docolon
        dw SHARPSCR,ONEPLUS,ONE,XDO
FORM1:  dw II,BUFFER,DROP,XLOOP,FORM1
        dw EXIT

; ----------------
;  THE INDEX WORD
; ----------------
;  ( n1 n2 ---)
;
; Shows the first lines of the screens n1 through n2
;
; : INDEX
;  1+ SWAP CR DO
;    I 2 .R               \ Print block number in field of 2 characters.
;    I BLOCK 1E TYPE      \ Print top line, only 30 characters of it.
;  LOOP ;
;
        head INDEX,5,"INDEX",docolon
        dw ONEPLUS,SWAP,XDO,CR
INDX1:  dw II,TWO,DOTR,II
        dw BLOCK,LIT,$1E,TYPE
        dw XLOOP,INDX1,EXIT

; ---------------
;  THE LOAD WORD
; ---------------
;  ( n -- )
;
; Begin interpretation of screen n by making it the input stream;
; preserve the locators of the present input stream (from >IN and BLK).
; If interpretation is not terminated explicitly it will be terminated
; when the input stream is exhausted. Control then returns to the input
; stream containing LOAD, determined by the input stream locators >IN
; and BLK.
;
; : LOAD
;   >IN @ >R BLK @ >R      \ Save old input stream on return stack.
;   BLK ! 0 >IN !
;   INTERPRET
;   R> BLK ! R> >IN ! ;    \ Restore old input stream.
;
        head LOAD,4,"LOAD",docolon
        dw TOIN,FETCH,TOR,BLK,FETCH,TOR
        dw BLK,STORE,ZERO,TOIN,STORE
        dw INTERPRET,FROMR,BLK,STORE,FROMR
        dw TOIN,STORE,EXIT

; --------------
;  THE RUN WORD
; --------------
;  ( -- )
;
; Get a file and load from it.
;
; : RUN   1 GET 1 LOAD ;
;
        head RUN,3,"RUN",docolon
        dw ONE,GET,ONE,LOAD,EXIT

; ----------------
;  THE LIST WORD
; ----------------
;  ( n -- )
;
; Prints the contents of screen n.
;
; : LIST DUP SCR ! CR ." Scr#"
;  DUP . CR BLOCK C/L 0 DO
;  DUP C/L TYPE
;  BEGIN ?TERMINAL 0= UNTIL
;  C/L + LOOP DROP ;
;
        head LIST,4,"LIST",docolon
        dw DUP,SCR,STORE,CR,XDQUOTE
        db $04,"SCR#"
        dw DUP,DOT,CR,BLOCK,CL,ZERO,XDO
LST1:   dw DUP,CL,TYPE
LST2:   dw QTERMINAL,ZEQUAL,QBRANCH,LST2
        dw CL,PLUS,XLOOP,LST1,DROP,EXIT

; ----------
;  ?LOADING
; ----------
;  ( -- )
;
; Reports an error if the current input stream isn't in the block buffer.
;
;    BLK @ 0= ABORT" NOT LOADING" ;
;
        nohead QLOADING,docolon
        dw BLK,FETCH,ZEQUAL,XABORTQ
        db 11,"NOT LOADING"
        dw EXIT

; ---------------
;  THE -->  WORD
; ---------------
;  ( -- )
;
; May only occur in a screen, causes the text interpreter to continue
; on the next screen.
;
; : -->
;   ?LOADING
;   1 BLK +! 0 >IN !
; ; IMMEDIATE
;
        immed NEXTSCREEN,3,"-->",docolon
        dw QLOADING
        dw ONE,BLK,PLUSSTORE
        dw ZERO,TOIN,STORE,EXIT

; ------------
;  THE ( WORD
; ------------
;  ( -- )                               "paren"
;
; Used in the form:
;      ( ccc)
; Accept and ignore comment characters from the input stream, until the
; next right parenthesis. As a word, the left parenthesis must be followed
; by one blank. It may freely be used while executing or compiling. An
; error condition exists if the input stream is exhausted before the right
; parenthesis.
;
; : (  29 WORD DROP ; IMMEDIATE
;
; Skip input until )
;
        immed PAREN,1,"(",docolon
        dw LIT,$29
        dw WORD
        dw DROP
        dw EXIT

; ------------
;  THE \ WORD
; ------------
;  ( -- )
;
; May only be used on a screen. Skips to the end of the current line.
; Used as a comment.
;
; : \   >IN @ C/L NEGATE AND C/L + >IN ! ; IMMEDIATE
;
        immed BKSLASH,1,'\',docolon
        dw QLOADING
        dw TOIN,FETCH,CL,NEGATE,AND_
        dw CL,PLUS,TOIN,STORE,EXIT

; -------
;  (CAT)
; -------
;  ( addr -- )
;
; Run time code for CAT
;
        nohead PCAT,docode
        MODULE PCAT
        push    de              ; save IP
        ld      a,(bc)          ; length is zero?
        or      a               ;
        jr      nz,L01

        ld      d,a             ; DE = 0, no path dir
        ld      e,a             ;
        jr      L02             ;

L01     ld      d,b             ; path/name address
        ld      e,c             ;
        call    fname2zx        ; convert ASCII chars to ZX chars
L02     call    opendir

L07     ld      e,22            ; prints 22 lines
L06     call    findnext        ; next dir entry
        cp      $3f             ; finished?
        jr      z,L03           ;

        call    prepbfrread
L05     ld      bc,$2007        ; BFR_READ
        in      a,(c)
        or      a               ; null terminated
        jr      z,L04
        ld      hl,ASCTAB       ; convert ZX char code to ASCII
        ld      bc,$003b        ;
        cpir                    ;
        ld      a,$5a           ;
        sub     c               ;
        call    prt_a           ; print char

        jr      L05
L04     ld      a,$0d           ; emit CR
        call    prt_a           ;
        dec     e
        jr      nz,L06

L08     call    KSCAN           ; after 22 entries, wait for a key press
        inc     L
        jr      z,L08

        ld      a,$7f
        in      a,($fe)         ; BREAK key?
        rra                     ;
        jr      c,L07           ; stop listing if BREAK key, otherwise continue

L03     pop     de              ; restore IP
        pop     bc              ; new TOS
        jp      (iy)
        ENDMODULE

; --------------
;  THE CAT WORD
; --------------
;  ( -- )
;
; Prints the directory list
;
; : CAT   BL WORD CR (CAT) CR ;
;
        head CAT,3,"CAT",docolon
        dw BL,WORD,CR,PCAT
        dw CR,EXIT

; ----------------
;  THE BLOAD WORD
; ----------------
;  ( addr -- )
;
; Load a file from sdcard to addr.
; Executed in the form:
;   addr BLOAD <name>
; LOAD the file <name> from sdcard and store it at addr.
;
; : BLOAD
;   BL WORD (BLOAD) ;
;
        head BLOAD,5,"BLOAD",docolon
        dw BL,WORD,bld1
bld1:   call    fname2zx        ; convert ASCII string to ZX string
        call    initfileoprd
        in      e,(c)           ; data length to DE
        in      d,(c)
        pop     hl              ; address to HL
        push    de              ; sv data lenght
        jr      ld_remaining

ld_readdata256:
        xor     a

ld_readdata:
        push    de
        ld      e,a             ; loop count
        ld      bc,0a007h
        out     (c),a           ; bytes to read into globalbuffer 0 = 256
        call    response40

   ; data is ready for us

        call  prepbfrread

ld_getbyte:
        in      a,(c)
        ld      (hl),a
        inc     hl
        dec     e
        jr      nz,ld_getbyte

ld_remaining:
        pop     de
        dec     d
        jp      p,ld_readdata256

ld_cleanup:
        ld      a,e             ; any stragglers?
        ld      e,00h
        or      a
        jr      nz,ld_readdata

close40:
        ld      bc,8007h        ; close file
        ld      a,80h
        out     (c),a
        call    response40
        call    slowfast
        pop     bc              ; adjust stack
        jp      EXIT

; ----------------
;  THE BSAVE WORD
; ----------------
;  ( addr u -- )
;
; Save u bytes from adr to sdcard
; Executed in the form:
;   addr u BSAVE <name>
; save at sdcard the contents of memory from addr to addr+u.
;
; : BSAVE
;   BL WORD (BSAVE) ;
;
        head BSAVE,5,"BSAVE",docolon
        dw BL,WORD,bsv1
bsv1:   call    fname2zx        ; convert ASCII string to ZX string
        ld      a,01h
        call    initfileop
        pop     de              ; data length to DE
        pop     hl              ; address to HL
        ld      a,d             ; null lenght?
        or      e               ;
        jr      z,close40       ; finish if so

        push    de              ; sv lenght
        jr      sv_remaining

sv_writedata256:
        xor     a
sv_writedata:
        push    de
        push    af
        ld      e,a             ; loop count
        call    prepbfrwrite

sv_getbyte:
        ld      a,(hl)
        out     (c),a
        inc     hl
        dec     e
        jr      nz,sv_getbyte
        pop     af
        ld      bc,0c007h
        out     (c),a           ; bytes to write to file from globalbuffer
        call    response40

sv_remaining:
        pop     de
        dec     d
        jp      p,sv_writedata256

sv_cleanup:
        ld      a,e             ; any stragglers?
        ld      e,00h
        or      a
        jr      nz,sv_writedata
        jr      close40

; ----------------
;  THE FSAVE WORD
; ----------------
;  ( -- )
;
; Save a copy of the user's vocabulary on the SD card.
; Executed in the form:
;   FSAVE <name.F79>
;
; : SAVE HERE FENCE !
;   4009 HERE DUP 4014 !
;   OVER - BSAVE ;
;
        head FSAVE,5,"FSAVE",docode
        push bc
        ld hl,(dp)
        ld (fence),hl
        ld ($4014),hl
        ld bc,$4009
        push bc
        or a
        sbc hl,bc
        ld b,h
        ld c,l
        jp BSAVE

; ----------------
;  THE FLOAD WORD
; ----------------
;  ( -- )
;
; Load an user's vocabulary from SD card.
; Executed in the form:
;   FLOAD <name.F79>
;
        head FLOAD,5,"FLOAD",docolon
        dw LIT,$4009,BLOAD,COLD

; -------------------
;  THE (FORGET) WORD
; -------------------
;  ( addr -- )
;
;  ( From ZX Spectrum Forth-83, by L. C. Benschop )
;
; : (FORGET)
;   ?fence              \ Check that we don't forget below FENCE
;   STOPOFF             \ Disable BREAK key while operation in progress
;   VOC-LINK @          \ Traverse all vocabularies.
;   BEGIN
;     ?DUP              \ While more vocabularies in list.
;   WHILE
;     2DUP U< IF
;       @ DUP VOC-LINK ! \ If vocabulary above forget address, unlink it
;                        \ from VOC-LINK list, discard vocabulary completely.
;     ELSE
;       DUP 8 -
;        BEGIN
;          @ 2- DUP 4 PICK U< \ Traverse vocabulary until below forget point
;        UNTIL
;       2+ OVER 8 - !         \ Update link pointer in vocabulary.
;       @                     \ Go to next vocabulary in VOC-LINK list.
;     THEN
;   REPEAT
;   DP !  STOPON ;            \ Update dictionary pointer.
;
        nohead PFORGET,docolon
        dw QFENCE,STOPOFF,VOC_LINK,FETCH
PFRGT1:
        dw QDUP,QBRANCH,PFRGT5,TWODUP,ULESS             ; BEGIN WHILE
        dw QBRANCH,PFRGT2,FETCH,DUP,VOC_LINK,STORE      ; IF
        dw BRANCH,PFRGT4                                ; ELSE
PFRGT2: dw DUP,LIT,8,MINUS
PFRGT3: dw FETCH,TWOMINUS,DUP,LIT,4,PICK,ULESS          ; BEGIN
        dw QBRANCH,PFRGT3,TWOPLUS,OVER,LIT,8,MINUS      ; UNTIL
        dw STORE,FETCH
PFRGT4: dw BRANCH,PFRGT1                                ; REPEAT
PFRGT5: dw DP,STORE,STOPON,EXIT

; -----------------
;  THE FORGET WORD
; -----------------
;  ( -- )
;
; Execute in the form:
;      FORGET  <name>
; Delete from the dictionary <name> and all words added to the dictionary
; after <name>, regardless of their vocabulary. Failure to find <name> or
; if <name> is below FENCE are error conditions.
;
; ( From ZX Spectrum Forth-83, by L. C. Benschop )
;
; : FORGET
;   [COMPILE] '          \ Find address of word to forget.
;   CONTEXT @ OVER U<
;   IF
;     [COMPILE] FORTH    \ Select FORTH vocabulary if CONTEXT is below forget
;                        \ address, so it will be removed completely.
;   THEN
;   CURRENT @ OVER U< IF
;     DEFINITIONS        \ Make sure CURRENT is not removed completely.
;   THEN
;   >NFA 2- (FORGET) ;
;
        head FORGET,6,"FORGET",docolon
        dw TICK,CONTEXT,FETCH,OVER,ULESS
        dw QBRANCH,FORG1,FORTH
FORG1:  dw CURRENT,FETCH,OVER,ULESS
        dw QBRANCH,FORG2,DEFINITIONS
FORG2:  dw NFA,TWOMINUS,PFORGET,EXIT

; ----------------
;  THE VLIST WORD
; ----------------
;  ( -- )
;
; List the word names of the CONTEXT vocabulary starting  with the most
; recent definition.
;
; : VLIST
;   CONTEXT @ @
;   BEGIN
;      2 SPACES DUP COUNT 1F AND TYPE
;      BEGIN ?TERMINAL 0= UNTIL
;      2- @ ?DUP 0=
;   UNTIL ;
;
        head VLIST,5,"VLIST",docolon
        dw CONTEXT,FETCH,FETCH
VLIST1: dw TWO,SPACES,DUP,COUNT,LIT,$1F,AND_
VLIST2: dw QTERMINAL,ZEQUAL,QBRANCH,VLIST2
        dw TYPE,TWOMINUS,FETCH,QDUP,ZEQUAL
        dw QBRANCH,VLIST1,EXIT

; ---------------
;  THE WARM WORD
; ---------------
;  ( -- )
;
; Warm start of Forth.
;
; : WARM
;   STOPOFF ' NOOP IS PAUSE usr_addr UP !
;   SP! DECIMAL  [COMPILE] FORTH DEFINITIONS
;   STOPON BLK @ ?DUP IF >IN @ THEN BELL QUIT ;
;
        head WARM,4,"WARM",docode
        ld      hl,FTFLAGS
        res     3,(hl)          ; STOPOFF
        ld      hl,NOOP
        ld      (PAUSE+1),hl
        ld      hl,usr_addr
        ld      (UP),hl
        call    docolon
        dw SPSTORE,DECIMAL,FORTH,DEFINITIONS,STOPON
        dw BLK,FETCH,QDUP,QBRANCH,WRM1,TOIN,FETCH
WRM1:   dw BELL,QUIT

; ----------------
;  THE ABORT WORD
; ----------------
;  ( -- )
;
; clears all stacks and restarts forth in a well-defined state by
; calling WARM.
;
        head ABORT,5,"ABORT",docode
        ld      hl,WARM
        jp      (hl)

; -------------------
;  THE (ABORT") WORD
; -------------------
;  ( f -- )
;
; Runtime part of ABORT" .  If flag is true, print the counted message.
; Then execute ABORT.
;
; : (ABORT")
;   R> COUNT ROT IF TYPE SPACE
;   HERE COUNT TYPE ABORT
;   ELSE + >R THEN ;
;
        nohead XABORTQ,docolon
        dw FROMR,COUNT,ROT,QBRANCH,XABOQ1
        dw TYPE,SPACE,HERE,COUNT,TYPE,ABORT
XABOQ1: dw PLUS,TOR
        dw EXIT

; -----------------
;  THE ABORT" WORD
; -----------------
;  ( f -- )                             "abort-quote"
;
; Used in a colon-definition in the form:
;           ABORT" stack empty"
; If the flag is true,  print the following text,  till ".
; Then execute ABORT.
;
; : ABORT"   COMPILE (ABORT") 22 WORD C@ 1+ ALLOT ; IMMEDIATE
;
        immed ABORTQUOTE,6,'ABORT"',docolon
        dw COMPILE,XABORTQ,LIT,$22,WORD
        dw CFETCH,ONEPLUS,ALLOT,EXIT

; ----------
;  (ERRNUM)
; ----------
;  ( f -- )
;
; Default code for ERRNUM.
;
; : (ERRNUM)   ABORT" ?" ;
;
        head PERRNUM,8,"(ERRNUM)",docolon
        dw XABORTQ
        db 1,"?"
        dw EXIT

; --------------
;  THE MEM WORD
; --------------
;  ( --  )
;
; Reports the free memory available
;
; : MEM   S0 @ 268 - HERE - U.
;   ." BYTES FREE." CR ;
;
        head FMEM,3,"MEM",docolon
        dw S0,FETCH,LIT,$10C,MINUS,HERE
        dw MINUS,UDOT,XDQUOTE
        db $0a,"BYTES FREE"
        dw EXIT

; ----------------------
;  THE 79-STANDARD WORD
; ----------------------
;  ( -- )                               79
; Informs that this is a Forth-79 system
;
; : 79-STANDARD
;   CR ." ZX81 FORTH-79"
;   CR ." TODDY SOFTWARE 2021"
;   CR SP@ HERE - 0C - U.
;   ." BYTES FREE." ;
;
        head STD79,11,"79-STANDARD",docolon
        dw CR,XDQUOTE
        db 19,"ZX81 FORTH-79 1.14",$0d
        dw XDQUOTE
        db 20,"TODDY SOFTWARE 2024",$0d
        dw FMEM,CR,EXIT

; ---------------
;  THE COLD WORD
; ---------------
;
; Cold start Forth system
;
        head COLD,4,"COLD",docode
        out     ($fd),a         ; disable NMI generator
        ld      hl,line1
        ld      (NXTLIN),hl
        ld      hl,(RAMTOP)
        ld      (lo),hl         ; sets the variable LO
        dec     hl
        ld      (limit),hl      ; Last byte of block buffer
        dec     h
        dec     h
        dec     h
        dec     h
        ld      (first),hl      ; First byte of block buffer
        dec     hl
        dec     hl
        ld      (RSP),hl        ; Set the return stack pointer
        ld      (r0),hl         ; Base of the return stack
        dec     h
        ld      (tib),hl        ; Terminal Input Buffer
        ld      sp,hl           ; Set the data stack pointer
        ld      (s0),hl         ; Base of the data stack
        ld      hl,PERRNUM
        ld      (ERRNUM+1),hl

        ld      a,$1e           ; default characters at ROM
        ld      i,a

        ld      ix,xisr         ; enable interrupt driver
        ld      iy,NEXT
        out     ($fe),a         ; enable NMI
        call    docolon
        dw STOPOFF,PAGE,DECIMAL
        dw FENCE,FETCH,PFORGET
        dw STD79,EMPTYBUF,WARM

; ---------------------
;  THE VOCABULARY WORD
; ---------------------
;
; : VOCABULARY
;   CREATE            \ Create mew word.
;   HERE 6 + ,        \ cell 1, contains latest def in this vocabulary,
;                     \         initialized to second dummy name field.
;   A081 ,            \ cell 2, first dummy name field, forms dummy
;                     \         header with cell 1 as link field, will be
;                     \         silently traversed by FIND.
;   CURRENT @ 2+ ,    \ cell 3 contains pointer to cell 2 in
;                     \        old CURRENT vocabulary,
;   A081 ,            \ cell 4  second dummy name field, forms dummy
;                     \        header with cell 3 as link field, will be
;                     \        silently traversed by FIND.
;   HERE VOC-LINK @ , VOC-LINK ! \ Link the new vocabulary in VOC-LINK list.
;                                \ cell 5 is link to previous vocabulary/
;   DOES>
;     CONTEXT !  ;    \ Set context VOCABULARY
;
        head VOCABULARY,10,"VOCABULARY",docolon
        dw CREATE,HERE,LIT,6,PLUS,COMMA
        dw LIT,$A081,COMMA,CURRENT,FETCH
        dw TWOPLUS,COMMA,LIT,$A081,COMMA
        dw HERE,VOC_LINK,FETCH,COMMA
        dw VOC_LINK,STORE
        dw XSCCODE
VOC1:   call dodoes
        dw CONTEXT,STORE,EXIT

; ----------------------
;  THE DEFINITIONS WORD
; ----------------------
;  ( --- )
;
; : DEFINITIONS   CONTEXT @ CURRENT ! ;
;
        head DEFINITIONS,11,"DEFINITIONS",docode
;        dw CONTEXT,FETCH,CURRENT,STORE
;        dw EXIT
	ld	hl,(context)
	ld	(current),hl
	jp	(iy)

; ---

RBRK	ld	a,(LAST_K)
	inc	a
	jr	nz,RBRK
	ld	a,$0f		; cursor off
	call	prt_a
	jp	ABORT

SPARE   ds $3c00-$

        assert  SPARE >= 0

; CHARACTER TABLE
;
CHARTAB
        INCBIN "Fonts/TF79B.FN"

        outend


;---------------------------------------------------------------------------------
; BASIC loader program ( Don't change it if you don't know what are you doing! )
;---------------------------------------------------------------------------------
;

                output TF79.P

                org     $4009

; SYSVARS which aren't saved

ERR_NR          equ     $4000
FLAGS           equ     $4001
ERR_SP          equ     $4002
RAMTOP          equ     $4004
MODE            equ     $4006
RSP             equ     $4007           ; Return Stack Pointer                 (old OLDPPC)

; SYSVARS which are. This is the start of the .P

VERSN:          db      0
UP              dw      usr_addr        ; User Area Pointer                    (old E_PPC)
D_FILE:         dw      dfile           ; Screen address
CURADD:         dw      dfile+1         ; Cursor address                       (old DF_CC)
LKEY:           db      0               ; Last key read                        (old VARS)
REP:            db      0               ; Counter for repeat keys
HCURADD:        dw      0               ; Cursor address on HR screen          (old DEST)
E_LINE:         dw      last
CH_ADD:         dw      last-1
X_PTR:          dw      0
STKBOT:         dw      last
STKEND:         dw      last
BERG:           db      0
MEM:            dw      0
FTFLAGS:        db      0               ; flags:  CFxxSKxA
                                        ;         C = cursor on/off, 1=on
                                        ;         F = Fast flag
                                        ;         S = Stop on/off
                                        ;         K = caps flag
                                        ;         A = attribute color mode, 1=active
                                        ;         x = flag not used

CMBDR           db      $01             ; Chroma color mode and border color   (old DF_SZ)
                                        ; +---+---+---+---+---+---+---+---+
                                        ; | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
                                        ; +---+---+---+---+---+---+---+---+
                                        ;   |   |   |   |   |   |   |   | 
                                        ;   |   |   |   |   |---+---+---+---+--- Border colour code
                                        ;   |   |   |   |
                                        ;   |   |   |   +---- Mode (0=Character code, 1=Attribute file).
                                        ;   |   |   +-------- 1=Enable color mode.
                                        ;   +---+------------ Reserved for future use (always set to 0).
ATTR            db      $1f             ; ink and paper colors attributes     (old S_TOP low)
IREG:           db      $1e 
LAST_K:         db      $FF,$FF
                db      $FF 
MARGIN:         db      55
NXTLIN:         dw      line1
OLDPPC:         dw      0
FLAGX:          db      0
INVCHR          db      0               ; character invert flag                (old STRLEN low)
CTIMER          db      22              ; timer for blinking cursor            (old STRLEN hidh)
T_ADDR:         dw      $0C8D
SEED:           dw      $EFC9
FRAMES:         dw      $FFFF
COORDS:         db      0,0
PR_CC:          db      $BC
S_POSN:         db      33,24
CDFLAG:         db      01000000B
PRTBUF:         ds      33

; ---------------------------------
;  Forth user and system variables
; ---------------------------------
; At MEMBOT area, so you can't use ROM calls that make use of this area.
; But who cares, you don't need it, Forth does everything for you ;-)
;
MEMBOT

usr_addr:
LINK    dw      LINK            ; LINK          \
        dw      0               ; 'SP            |
s0      dw      0               ; S0              > Per task variables
r0      dw      0               ; R0             |
        dw      10              ; BASE          /

dp      dw      last            ; DP            \
fence   dw      last            ; FENCE          |
state   dw      0               ; STATE          |
        dw      0               ; >IN            |
hld     dw      0               ; HLD            |
        dw      forth           ; VOC-LINK        > Global variables
context dw      0               ; CONTEXT        |
current dw      0               ; CURRENT        |
dpl     dw      0               ; DPL            |
        dw      0               ; BLK            |
        dw      0               ; SCR           /



;= First BASIC line, calls ASM ==================================

line1:
   db $00,$01           ; line number
   dw line2-$-2         ; line length
   db $e2,$0b,$32       ; CONFIG "M=L"  ; Sets RAM to 8K-40K
   db $14,$31,$0b       ;
   db $76               ; N/L

line2:
   db $00,$02           ; line number
   dw line3-$-2         ; line length
   db $ef,$0b,$39,$2b   ; LOAD "TF79.BIN;8192"
   db $23,$25,$1b,$27   ;
   db $2e,$33,$19,$24   ;
   db $1d,$25           ;
   db $1e,$0b           ;
   db $76               ; N/L

dblINI  equ (4*COLD)& $7FFF    ; valid for addresses from $2000-$3fff

line3:
   db $00,$03           ; line number
   dw dfile-$-2         ; line length
   db $f9,$d4,$1c       ; RAND USR 0
   db $7e,$8e           ; FP value
   db dblINI/256        ;
   db dblINI&255        ;
   db $00,$00           ;
   db $76               ; N/L

;- Display file --------------------------------------------

dfile:
        db $76
 .24    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,$76

;- BASIC-Variables ----------------------------------------
; Forth don't use it, byte preserved only to separate DFile from user area
;
DF_END:
   dw   $ff             ; dfile end mark

;- End of BASIC area ----------------------------

; ----------------
;  THE FORTH WORD
; ----------------
;
        immed FORTH,5,"FORTH",docode
        call    VOC1            ; Create code field, CALL to DOES> part of VOCABULARY.
        dw      link            ; cell 1 latest def in this vocabulary
        dw      $a081           ; cell 2 First dummy name field
        dw      0               ; cell 3 link to old CURRENT, null for FORTH.
        dw      $a081           ; cell 4 Second dummy name field.
forth:  dw      0               ; Mark address to patch VOC-LINK, store
                                ; null pointer in cell 5.
; ---

last:   ; where user dictionary starts

        outend

        INCLUDE EXPORT.INC

; Show code statistics when compiling

spare       equ $3c00-SPARE
code_lenght equ last-VERSN

        display "Spare bytes: ",/D,spare," bytes"
        display "Program lenght: ",/D,code_lenght," bytes"
        display "COLD address = ",/a,COLD
